//
//  cellular_automata.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#include <stdio.h>
#include <stdlib.h>
#include "cellular_automata.h"
#include "genotype.h"
#include "phenotype.h"


extern setting_flags flags;


void free_automata(automata * ca){
    free_genotype(ca->gtype);
    free(ca->ptype);
    free(ca);
}

automata * init_automata(ca_values values){

    automata * ca = malloc(sizeof(automata));
    
    ca->val.grid_size = values.grid_size;
    ca->val.dim = values.dim;
    ca->val.states = values.states;
    ca->nh_size = ca->val.dim == 1 ? 3 : 5;
    return ca;
}


automata * init_random_automata_with(ca_values values, int buffersize){
    automata * ca = init_automata(values);
    
    ca->ptype = init_empty_ptype(ca);
    ca->gtype = init_full_genotype(ca, NEXT_VALUE);
    setup_ptype(ca->ptype, buffersize);
    
    set_initial_condition_on_ptype(ca->ptype);
    
    return ca;
}

automata * init_restricted_automata_with(ca_values values, int buffersize, int restriction){
    automata * ca = init_automata(values);
    
    ca->ptype = init_empty_ptype(ca);
    if (flags.use_zero_init){
        ca->gtype = init_restircted_genotype_empty_genes(restriction);
    }else{
        ca->gtype = init_restricted_genotype(restriction, values.states, RANDOM);
    }
    setup_ptype(ca->ptype, buffersize);
    
    set_initial_condition_on_ptype(ca->ptype);
    
    return ca;
}

automata * init_growing_automata_with(ca_values values, int buffersize, int start_size){
    automata * ca = init_automata(values);
    
    ca->ptype = init_empty_ptype(ca);
    ca->gtype = init_empty_genotype(start_size);

    setup_ptype(ca->ptype, buffersize);
    
    set_initial_condition_on_ptype(ca->ptype);
    
    return ca;
}