//
//  development.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#include <stdio.h>
#include <math.h>
#include <pthread.h>
#include "development.h"
#include "genotype.h"
#include "phenotype.h"
#include "file_helper.h"
#include "attractor_checking.h"
#include "bucket_sort.h"
#include "timeing.h"
#include "visualization.h"

static long development_step = 1;
static int excessive_storing = 0;
excess * storage = NULL;
int storage_size = 0;
static int max_storage = 2000;
void set_excessive(int a){
    excessive_storing = a;
    if (excessive_storing){
        storage = calloc(max_storage, sizeof(excess));
        storage_size = 0;
    }else {
        free(storage);
        storage = NULL;
        storage_size = 0;
    }
}

void grow_excessive(){
    max_storage = max_storage * 2;
    storage = realloc(storage, sizeof(excess) * max_storage);
}




void perform_excessive_storing(automata * ca){
    if (max_storage < storage_size){
        grow_excessive();
    }
    
    for (int i = 0; i < ca->ptype->buffer_size; i++){
        char current = ca->ptype->crnt->buffer[i];
        char next = ca->ptype->nxt->buffer[i];
        
        if (current == next){//nochage
            storage[storage_size].unchange ++;
        }
        else if (next == 0){ //death
            storage[storage_size].death ++;            
        }
        else if (current == 0){ //grow
            storage[storage_size].life ++;                    
        }
        else{ //diff
            storage[storage_size].change ++;                                
        }

    }
    storage[storage_size].mods = ca->gtype->mods;
    ca->gtype->mods = 0;
    storage_size ++;
}

typedef struct step_args{
    automata * ca;
    int height; 
    int width; 
    int total_size;
    int bits_n;
    uint32_t mask; 
    uint32_t value;
    int from;
    int to;
}step_args;


void *preform_step(step_args * sa){
    int x = 0;
    int y = 0;
    
    for (int i = sa->from; i < sa->to; i++){
        x = i % sa->ca->ptype->buffer_width;
        y = floor(i / sa->ca->ptype->buffer_width);
        
        uint32_t gene = 0;
        char * ptr = sa->ca->ptype->crnt->buffer;
        /* Next, this is added so that in the case of no rules found
         * the site can retain its value.
         */
        gene |= (ptr[i] & sa->mask);
        
        //East
        int tmp_poss = x == sa->width - 1 ? i - sa->width + 1 : i + 1;
        gene |= (ptr[tmp_poss] & sa->mask) << (sa->bits_n * 1);
        
        //Center
        gene |= (ptr[i] & sa->mask) << (sa->bits_n * 2);
        
        //West
        tmp_poss = x == 0 ? i + sa->width - 1 : i - 1;
        
        gene |= (ptr[tmp_poss] & sa->mask) << (sa->bits_n * 3);
        if (sa->ca->val.dim == 2){
            //South
            tmp_poss = y == sa->height - 1 ? i + sa->width - sa->total_size : i + sa->width ;
            gene |= (ptr[tmp_poss] & sa->mask) << (sa->bits_n * 4);
            //North
            tmp_poss = y == 0 ? i - sa->width + sa->total_size : i - sa->width ;
            gene |= (ptr[tmp_poss] & sa->mask) << (sa->bits_n * 5);
        }
        
        sa->ca->ptype->nxt->buffer[i] = get_next_state_apply_rules_to_gene(sa->ca->gtype, gene, sa->mask, sa->ca->val.states);

    }
}



uint16_t step(automata * ca){
    /*
     Finn ut hvilket buffer som blir indikert, og hvilket som burde vært indikert... jobb derifra
     */
   /* step_args args1, args2;
    args1.ca = ca;
    args1.height = ca->ptype->buffer_size / ca->ptype->buffer_width;
    args1.width = ca->ptype->buffer_width;
    args1.total_size = args1.height*args1.width;
    args1.bits_n = (int)ceil((double)ca->val.states / 2.0);
    args1.mask = pow(2, args1.bits_n) -1;
    args1.value = 0;
    args2 = args1;
    
    pthread_t job1, job2;
    args1.from = 0;
    args1.to = ceil(ca->ptype->buffer_size/2);
    args2.from = ceil(ca->ptype->buffer_size/2);
    args2.to = ca->ptype->buffer_size;
    pthread_create(&job1, NULL, &preform_step, &args1);
    pthread_create(&job2, NULL, &preform_step, &args2);
    
    pthread_join(job1, NULL);
    pthread_join(job2, NULL);
    */
    
    

    int height = ca->ptype->buffer_size / ca->ptype->buffer_width;
    int width = ca->ptype->buffer_width;
    int total_size = height*width;
    int bits_n = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_n) -1;
    uint32_t value = 0;
    
    
    int x = 0;
    int y = 0;
    
    for (int i = 0; i < ca->ptype->buffer_size; i++){
        x = i % ca->ptype->buffer_width;
        y = floor(i / ca->ptype->buffer_width);
        
        uint32_t gene = 0;
        char * ptr = ca->ptype->crnt->buffer;
        /* Next, this is added so that in the case of no rules found
         * the site can retain its value.
         */
        gene |= (ptr[i] & mask);
        
        //East
        int tmp_poss = x == width - 1 ? i - width + 1 : i + 1;
        gene |= (ptr[tmp_poss] & mask) << (bits_n * 1);
        
        //Center
        gene |= (ptr[i] & mask) << (bits_n * 2);
        
        //West
        tmp_poss = x == 0 ? i + width - 1 : i - 1;
        
        gene |= (ptr[tmp_poss] & mask) << (bits_n * 3);
        if (ca->val.dim == 2){
            //South
            tmp_poss = y == height - 1 ? i + width - total_size : i + width ;
            gene |= (ptr[tmp_poss] & mask) << (bits_n * 4);
            //North
            tmp_poss = y == 0 ? i - width + total_size : i - width ;
            gene |= (ptr[tmp_poss] & mask) << (bits_n * 5);
        }
        
        ca->ptype->nxt->buffer[i] = get_next_state_apply_rules_to_gene(ca->gtype, gene, mask, ca->val.states);
        
    }
    
    
    if (excessive_storing){
        perform_excessive_storing(ca);
    }
    return value;
}



uint16_t phantom_step(automata * ca){
    /*
     Finn ut hvilket buffer som blir indikert, og hvilket som burde vært indikert... jobb derifra
     */
    int height = ca->ptype->buffer_size / ca->ptype->buffer_width;
    int width = ca->ptype->buffer_width;
    uint32_t value = 0;
    for (int i = 0; i < height; i++){
        for (int j = 0; j < width; j++){
            int current_poss = i * width + j;
            
            uint16_t next = ca->ptype->crnt->buffer[current_poss];
            
            next = rol(next, i+5);
            value |= next;
             
        }
    }
    return value;
}



int array_cmp(char * a, char * b, int length){

    for (int i = 0; i < length; i++){
        if (a[i] != b[i]){
            return 0;
        }
    }
    return 1;
}

long check_ptype_vs_buffer(linked_buffer * ptype, linked_buffer * buffer, int size, int number){
    for (int i  = 0; i < number; i++){
        if (array_cmp(ptype->buffer, buffer->buffer, size)){
            return buffer->step;
        }
        buffer = buffer->next;
    }
    return -1;
}
static int checker = 0;
int find_attractor_bucket(automata * ca, attractor * a, int number_of_mem_checks, int use_temp){
    
    int filled_buffers = 0;
    int number_of_mem_buffers = ca->ptype->nr_buffs;
    
    linked_buffer * oldest_mem_buffer = get_indicated_mem_buffer(ca->ptype);
    linked_buffer * check_with = NULL;
    linked_buffer * check_against = NULL;

    int reverses = number_of_mem_buffers - number_of_mem_checks + 1;
    reverses = fmod(reverses, number_of_mem_buffers);
    
    for (int i = 0; i < reverses; i++){
        oldest_mem_buffer = oldest_mem_buffer->next;
    }
    checker = 0;
    //check against data on HD
    if (development_step > ca->ptype->nr_buffs){

        check_with = oldest_mem_buffer;
        for (int i = 0; i < number_of_mem_checks; i++){
            if (checker == 310 || checker == 856){
                checker = checker;
            }
            checker ++;
            open_read_file(use_temp);
            bucket * relevant_bucket = get_relevant_bucket(check_with->key);
            content * t = relevant_bucket->contents;
            if (t == NULL){
                t = NULL;
            }
            while (t != NULL){
                filled_buffers = get_buffers_from(t->step);
                if (filled_buffers != 1){
                    t = t->nxt;
                    break;
                }
            
                if (array_cmp(check_with->buffer, get_fist_hd_buffer()->buffer, ca->ptype->buffer_size) ){
                    a->transient = (int)get_fist_hd_buffer()->step ;
                    a->attractor_length = (int)check_with->step - a->transient;
                    a->trajectory = a->transient + a->attractor_length;
                    close_read_file(use_temp);
                    return 1;
                }
            
                t = t->nxt;
            }            
            
            close_read_file(use_temp);
            check_with = check_with->next;
        }
    }

    //check against current deveopment loop
    check_against = oldest_mem_buffer;
    check_with = oldest_mem_buffer->next;
    
    for(int i  = 1; i < number_of_mem_checks; i++){

        for (int j = 0; j < i; j++){
            
            if (check_against->key != check_with->key){
                check_against = check_against->next;
                continue;
            }
            
            if (array_cmp(check_with->buffer, check_against->buffer, ca->ptype->buffer_size)){
                a->transient = (int)check_against->step ;
                a->attractor_length = (int)check_with->step - a->transient;
                a->trajectory = a->transient + a->attractor_length;
                
                return 1;   
            }
            check_against = check_against->next;
        }
        check_against = oldest_mem_buffer;
        check_with = check_with->next;
    }
    return 0;
}

int find_attractor_bruteforce(automata * ca, attractor * a, int number_of_mem_checks, int use_temp, const fpos_t * current_pos, int needio){
    
#if TIMING == 1
    start_timeing();
#endif
    
    int filled_buffers = 0;
    int number_of_mem_buffers = ca->ptype->nr_buffs;
    
    linked_buffer * oldest_mem_buffer = get_indicated_mem_buffer(ca->ptype);
    linked_buffer * check_with = NULL;
    linked_buffer * check_against = NULL;
    
    int reverses = number_of_mem_buffers - number_of_mem_checks + 1;
    reverses = reverses % number_of_mem_buffers;
    
    for (int i = 0; i < reverses; i++){
        oldest_mem_buffer = oldest_mem_buffer->next;
    }
    
    //check against data on HD
    open_read_file(use_temp);
    
    int checked_so_far = 0;
#if TIMING == 1
    save_timeing_to(ATR);
#endif
    while (needio && ( filled_buffers = fill_hd_buffer(current_pos) ) > 0 ){
#if TIMING == 1
        start_timeing();
#endif
        check_with = get_fist_hd_buffer();
        
        for (int i = 0; i < filled_buffers; i++){
            long hit_step = check_ptype_vs_buffer(check_with, oldest_mem_buffer, ca->ptype->buffer_size, number_of_mem_checks);
            
            if (hit_step >= 0){
                a->transient = (int)check_with->step;
                a->attractor_length =  (int)(hit_step - a->transient);
                a->trajectory = a->transient + a->attractor_length;

                close_read_file(use_temp);
#if TIMING == 1
                save_timeing_to(ATR);
#endif
                return 1;
            }
            check_with = check_with->next;
        }
        checked_so_far += filled_buffers;
#if TIMING == 1
        save_timeing_to(ATR);
#endif
    }
    close_read_file(use_temp);
    
#if TIMING == 1
    start_timeing();
#endif
    //check against current deveopment loop
    check_against = oldest_mem_buffer;
    check_with = oldest_mem_buffer->next;

    for( int i = 1; i < number_of_mem_checks; i++){
        long hit_number = check_ptype_vs_buffer(check_with, check_against, ca->ptype->buffer_size, i );
        if (hit_number >= 0){
            a->transient = (int)hit_number;
            a->attractor_length = (int)(check_with->step - a->transient);
            a->trajectory = a->transient + a->attractor_length;
            if (a->attractor_length == 4){
                a->attractor_length = 4;
            }
#if TIMING == 1
            save_timeing_to(ATR);
#endif
            return 1;
        }
        check_with = check_with->next;
    }
#if TIMING == 1
    save_timeing_to(ATR);
#endif
    return 0;
}



attractor * run_development(automata * ca, int steps_n, int use_temp){

#if TIMING == 1
    start_timeing();
#endif
    
    int i = 0;
    int saved = 0;
    int did_find = 0;
    int buffer_size = ca->ptype->nr_buffs;
    attractor * a = (attractor*)malloc(sizeof(attractor));
    
#if USE_ATTRACTOR_FORCE == 0
    init_butckets_(UINT16_MAX);
#endif
    
    set_initial_condition_on_ptype(ca->ptype);
    
    uint16_t key = phantom_step(ca);
    ca->ptype->crnt->key = key;
    
    ca->ptype->crnt->step = development_step -1;

#if USE_ATTRACTOR_FORCE == 0
    add_to_bucket(ca->ptype->crnt->key, development_step - 1);
#endif
    fpos_t file_pos = 0;
//write seperator
    clear_tmp();
    
    if (!use_temp){
        open_save_file(use_temp);
        file_pos = get_file_pos();
        write_deveopment_start();
        close_save_file(use_temp);
    }
    
#if TIMING == 1
    save_timeing_to(DEV);
#endif
    //reset the buffer for used chromosomes.
    for (int i = 0; i < ca->gtype->size; i++){
        ca->gtype->used[i] = 0;
    }
    ca->gtype->mods = 0;
    
    while(i < steps_n && !did_find){
        
#if TIMING == 1
        start_timeing();
#endif
        saved = 0;
        
        uint16_t key = step(ca);
        
        shift_ptype_buffer(ca);
        ca->ptype->crnt->key = key;
        ca->ptype->crnt->step = development_step;

#if USE_ATTRACTOR_FORCE == 0
        add_to_bucket(key, development_step );
#endif
        
        development_step ++;
        
      
#if TIMING == 1
        save_timeing_to(DEV);
#endif
        
        if (development_step > 0 && (development_step % buffer_size) == 0){
//check against tmp file

#if USE_ATTRACTOR_FORCE == 1
            did_find = find_attractor_bruteforce(ca, a, buffer_size, use_temp, &file_pos, buffer_size < development_step  );
#else
            did_find = find_attractor_bucket(ca, a, buffer_size, use_temp);
#endif


            saved = 1;
            if (did_find){
                long long tmp = development_step - (development_step - a->trajectory) + 1;
                tmp += buffer_size;
                int save_amount = tmp % buffer_size;
                
                if (save_amount > 0){
//write to tmp file
#if TIMING == 1
                    start_timeing();
#endif
                    save_number_mem_buffer_to_disk(ca, 0, save_amount, use_temp);
#if TIMING == 1
                    save_timeing_to(IOW);
#endif
                    break;
                }
            }
//write to tmp file
#if TIMING == 1
            start_timeing();
#endif
            save_mem_buffer_to_disk(ca, use_temp);
#if TIMING == 1
            save_timeing_to(IOW);
#endif
        }
        i++;
    }

    if (!saved && !did_find){
        int save_amount = development_step % buffer_size;
        int skip_amount = buffer_size - save_amount;

//check against tmp file
#if USE_ATTRACTOR_FORCE == 1
        did_find = find_attractor_bruteforce(ca, a, save_amount, use_temp, &file_pos, buffer_size < development_step );
#else
        did_find = find_attractor_bucket(ca, a, save_amount, use_temp);
#endif
//write to tmp file
#if TIMING == 1
        start_timeing();
#endif
        save_number_mem_buffer_to_disk(ca, skip_amount, save_amount, use_temp);
#if TIMING == 1
        save_timeing_to(IOW);
#endif
    }
    if (!did_find){
        free(a);
        a = NULL;
    }
    development_step = 1;
#if USE_ATTRACTOR_FORCE == 0
    free_buckets();
#endif
    
    return a;
}






