//
//  experimet.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "experimet.h"

#include "file_helper.h"
#include "gplotilizer.h"
#include "genetic_algorithm.h"

#include "tweaking.h"

#include "population.h"
#include "attractor_checking.h"
#include "mersenne_twister.h"


//Extern set
int rounds = 0;
int nr_of_states = 3;
int size_of_grid = 4;
int nr_of_dimensions = 2;

//Extern get
extern setting_flags flags;
extern int max_step;

extern int nr_of_success;
extern int wanted_attractor;

extern int nr_of_changed;
extern int nr_of_unchanged;
extern double mutated_bits_pr_individ;
extern int nr_of_elites;
extern double fitness_activation_ratio;

void find_custom_attractor(population * p, int relative_close, int add_tresh, int add_limit){
    double fit_demand = 1.00;

    init_genetic_algorithm_with_add_gene(fit_demand, p, wanted_attractor, relative_close, add_tresh, add_limit);
    
    int max_number_of_loops = max_step;
    int best_fit = algorithm_loop(max_number_of_loops);
    
    nr_of_success += flags.is_successfull;
    
    save_stats_for_population(p, best_fit);
    free_genetic_algorithm();
}

void find_attractor(population * p, int add_tresh, int add_limit){
    int realtive_close = 0;
    nr_of_changed = 0;
    nr_of_unchanged = 0;

    switch (wanted_attractor) {
        case 10:
            realtive_close  = RELATIVE_CLOSE_10;
            break;
            
        case 40:
            realtive_close  = RELATIVE_CLOSE_40;
            break;
            
        case 100:
            realtive_close  = RELATIVE_CLOSE_100;
            break;
            
        case 200:
            realtive_close  = RELATIVE_CLOSE_200;
            break;
            
        case 500:
            realtive_close  = RELATIVE_CLOSE_500;
            break;
            
        case 1500:
            realtive_close  = RELATIVE_CLOSE_1500;
            break;
            
        default:
            printf("No function for such attractor length, defaulting relative close to 1500\n");
            realtive_close  = RELATIVE_CLOSE_1500;
            break;
    }
    find_custom_attractor(p, realtive_close, add_tresh, add_limit);
}


void full_genotype_attractor_size(){

    int pop_size = POPULATION_SIZE;
    int number_of_mem_buffs_pr_ca = MEM_BUFFERS;
    int phenotype_hd_buffer_size  = nr_of_dimensions == 1 ? size_of_grid : size_of_grid * size_of_grid;
    
    population * p = init_population(pop_size,
                                     (ca_values){.grid_size=size_of_grid,
                                         .states=nr_of_states,
                                         .dim=nr_of_dimensions},
                                     number_of_mem_buffs_pr_ca);
    
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    find_attractor(p, -1, 0);
    free_hd_buffers();
}

void restricted_genotype_attractor_size(int restriction_size){

    int gtype_restriction = restriction_size;
    
    int pop_size = POPULATION_SIZE;
    int number_of_mem_buffs_pr_ca = MEM_BUFFERS;
    int phenotype_hd_buffer_size  = nr_of_dimensions == 1 ? size_of_grid : size_of_grid * size_of_grid;
    
    population * p = init_restricted_population(pop_size,
                                                (ca_values){.grid_size=size_of_grid,
                                                    .dim=nr_of_dimensions,
                                                    .states=nr_of_states},
                                                number_of_mem_buffs_pr_ca,
                                                gtype_restriction);
    
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    find_attractor(p, -1, 0);
    free_hd_buffers();
}

void growing_genotype_attractor_size(int start_size, int add_tresh){

    int pop_size = POPULATION_SIZE;
    int number_of_mem_buffs_pr_ca = MEM_BUFFERS;
    int phenotype_hd_buffer_size  = nr_of_dimensions == 1 ? size_of_grid : size_of_grid * size_of_grid;
    
    population * p = init_growing_population(pop_size,
                                             (ca_values){.grid_size=size_of_grid,
                                                 .dim=nr_of_dimensions,
                                                 .states=nr_of_states},
                                             number_of_mem_buffs_pr_ca,
                                             start_size);
    
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    find_attractor(p, add_tresh, ADD_LIMIT);
    free_hd_buffers();
}

void run_regulate_testing(int repetitions, GENOMTYPE type,int restrict_value){
    for (int i  = 0; i < repetitions; i++){
        switch (type) {
            case GROWING:
                growing_genotype_attractor_size(1, restrict_value);
                break;
            case RESTRICTED:
                restricted_genotype_attractor_size(restrict_value);
                break;
            case FULL:
            default:
                full_genotype_attractor_size();
                break;
        }
    }
}

double find_best_mutation_rate(const char * surfix){
    
    double mutation_rate[4] = {0.6, 0.05, 0.1, 0.5};
    double average[4] = {0.0};

    char * filenames[4] = {0};
    char fn1[256] = {0};
    strcpy(fn1, "mut_0.01");
    strcat(fn1, surfix);
    filenames[0] = fn1;
    
    char fn2[256] = {0};
    strcpy(fn2, "mut_0.05");
    strcat(fn2, surfix);
    filenames[1] = fn2;

    char fn3[256] = {0};
    strcpy(fn3, "mut_0.10");
    strcat(fn3, surfix);
    filenames[2] = fn3;

    char fn4[256] = {0};
    strcpy(fn4, "mut_0.50");
    strcat(fn4, surfix);
    filenames[3] = fn4;
    
    
    for (int i = 0; i < 4; i++){
        set_filename(filenames[i], 1);
        mutated_bits_pr_individ = mutation_rate[i];

        
        open_save_file(0);
        write_regulation_heading(RESTRICT_LOW);
        close_save_file(0);

        init_genrand(GLOBAL_RANDOM_SEED);
        
        run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW*2);
        
        average[i] = (double)nr_of_success / rounds;
        
        open_attractor_dev_file();
        write_success_rate(rounds);
        close_attractor_dev_file();
        
        nr_of_success = 0;
    }
        
    int highest = -1;
    for (int i = 0; i < 4; i++){
        if (highest < average[i])
            highest = i;
    }
    return mutation_rate[highest];
}

int find_best_elitism_size(int mutation_rate){
    int elites_array[3] = {0, 1, 10};
    double average[3] = {0.0};
    char * filenames[3] = {"elitism_0", "elitism_1", "elitism_10"};
    
    for (int i = 0; i < 3; i++){
        set_filename(filenames[i], 1);
        nr_of_elites = elites_array[i];
        
        open_save_file(0);
        write_regulation_heading(RESTRICT_LOW);
        close_save_file(0);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW);
        
        average[i] = (double)nr_of_success / rounds;
        
        open_attractor_dev_file();
        write_success_rate(rounds);
        close_attractor_dev_file();
        
        nr_of_success = 0;
    }
    
    int highest = -1;
    for (int i = 0; i < 3; i++){
        if (highest < average[i])
            highest = i;
    }
    return elites_array[highest];
}

int find_best_fitness_function(char * mut_rate){
    double average[8] = {0.0};
    double fitness_vals[4] = {0.05, 0.10, 0.20, 0.35};
    char * filenames[8] = {0};
    char fn1[256] = {0};
    strcpy(fn1, "fitness_expon_05");
    strcat(fn1, mut_rate);
    filenames[0] = fn1;

    char fn2[256] = {0};
    strcpy(fn2, "fitness_expon_10");
    strcat(fn2, mut_rate);
    filenames[1] = fn2;
    
    char fn3[256] = {0};
    strcpy(fn3, "fitness_expon_20");
    strcat(fn3, mut_rate);
    filenames[2] = fn3;
    
    char fn4[256] = {0};
    strcpy(fn4, "fitness_expon_35");
    strcat(fn4, mut_rate);
    filenames[3] = fn4;
    
    char fn5[256] = {0};
    strcpy(fn5, "fitness_lin_05");
    strcat(fn5, mut_rate);
    filenames[4] = fn5;
    
    char fn6[256] = {0};
    strcpy(fn6, "fitness_lin_10");
    strcat(fn6, mut_rate);
    filenames[5] = fn6;
    
    char fn7[256] = {0};
    strcpy(fn7, "fitness_lin_20");
    strcat(fn7, mut_rate);
    filenames[6] = fn7;
    
    char fn8[256] = {0};
    strcpy(fn8, "fitness_lin_35");
    strcat(fn8, mut_rate);
    filenames[7] = fn8;
    
    //run weighttesting for exponential fitness function
    flags.use_linear_fitness = 0;
    
    for (int i = 0; i < 4; i++){
        set_filename(filenames[i], 1);
        fitness_activation_ratio = fitness_vals[i];
    
        open_save_file(0);
        write_regulation_heading(RESTRICT_LOW);
        close_save_file(0);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW);
        
        average[i] = (double)nr_of_success / rounds;
        
        open_attractor_dev_file();
        write_success_rate(rounds);
        close_attractor_dev_file();
        
        nr_of_success = 0;
    }

    
    //run weighttesting for linear fitness function
    flags.use_linear_fitness = 0;
    
    for (int i = 0; i < 4; i++){
        set_filename(filenames[4 + i], 1);
        fitness_activation_ratio = fitness_vals[i];
        
        open_save_file(0);
        write_regulation_heading(RESTRICT_LOW);
        close_save_file(0);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW);
        
        average[4 + i] = (double)nr_of_success / rounds;
        
        open_attractor_dev_file();
        write_success_rate(rounds);
        close_attractor_dev_file();
        
        nr_of_success = 0;
    }
    
    
    int highest = -1;
    for (int i = 0; i < 8; i++){
        if (highest < average[i])
            highest = i;
    }
    return highest;
}


void adjust(){
    rounds = 20;                        //Number of rounds pr setting
    flags.use_zero_init = 0;                  //Genotype initiation type
    nr_of_elites = 0;                   //Number of elites
    fitness_activation_ratio = 0;       //Starting with 0 weighting on active rules
    wanted_attractor = WANTED_ATTRACTOR_40;
    flags.use_linear_fitness = 1;
    double m_rate_lin = find_best_mutation_rate("_lin");

    flags.use_linear_fitness = 0;
    double m_rate_exp = find_best_mutation_rate("_exp");
    
    
    double m_rate_max =  m_rate_exp;
    if (m_rate_exp < m_rate_lin){
        m_rate_max = m_rate_lin ;
        flags.use_linear_fitness = 1;
    }
    mutated_bits_pr_individ = m_rate_max;//set the mutation rate based of the genomesize
    
    printf("\nbest_mutation_rate == %f\n\n", m_rate_max);
   
    int best_e_rate = find_best_elitism_size(m_rate_max);
    nr_of_elites = best_e_rate;
    
    int fitness_index = find_best_fitness_function("");
    set_filename("settings_result", 1);
    
    open_save_file(0);
    write_param_results(m_rate_max, best_e_rate, fitness_index);
    close_save_file(0);

    printf("\nfinished adjustion\n");
}

void growth_test(){
    rounds = 20;                        //Number of rounds pr setting
    flags.use_zero_init = 1;                  //Genotype initiation type
    flags.use_linear_fitness = 1;
    nr_of_elites = 0;                   //Number of elites
    mutated_bits_pr_individ = 0.6;
    fitness_activation_ratio = 0.2;       //Starting with 0 weighting on active rules
    wanted_attractor = WANTED_ATTRACTOR_40;
    
    set_filename("growth_test", 1);
    
    open_save_file(0);
    write_regulation_heading(RESTRICT_LOW);
    close_save_file(0);
    
    init_genrand(GLOBAL_RANDOM_SEED);
    run_regulate_testing(rounds, GROWING, RESTRICT_LOW);
    
    
    open_attractor_dev_file();
    write_success_rate(rounds);
    close_attractor_dev_file();
}

void all_test_same_setup(){
    rounds = 20;                        //Number of rounds pr setting
    wanted_attractor = WANTED_ATTRACTOR_40;
    flags.use_zero_init = 1;                  //Genotype initiation type
    flags.use_linear_fitness = 1;
    fitness_activation_ratio = 0.15;
    
    int elites[2] = {0, 1};
    double mutation[4] = {0.8, 0.9, 1.0, 1.1};
    char * filename_full[8] = {"fu_e0,m0.8","fu_e0,m0.9","fu_e0,m1.0","fu_e0,m1.1","fu_e1,m0.8","fu_e1,m0.9","fu_e1,m1.0","fu_e1,m1.1"};
    
    for (int i = 0; i < 2; i ++){
        nr_of_elites = elites[i];
        for (int j = 0; j < 4; j++){
            mutated_bits_pr_individ = mutation[(i*4 + j)];

            //full
            
            set_filename(filename_full[(i*4 + j)], 1);
            
            open_save_file(0);
            write_regulation_heading(RESTRICT_LOW);
            close_save_file(0);
            
            init_genrand(GLOBAL_RANDOM_SEED);
            run_regulate_testing(rounds, FULL, RESTRICT_LOW);
            
            
            open_attractor_dev_file();
            write_success_rate(rounds);
            close_attractor_dev_file();
            
        }
    }
}


void run_multiple_experiments(){

/*
    max_step = 1500;
    all_test_same_setup();
    growth_configure();
    zero_init_restricted_exponential();
    zero_init_restricted_linear();
    growth_test();
    adjust();
*/
    
    

}
