//
//  file_helper.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/29/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <math.h>
#include <string.h>
#include "file_helper.h"
#include "phenotype.h"
#include "attractor_checking.h"
#include "genetic_algorithm.h"
#include "timeing.h"
#include "stats.h"



static FILE * usage_file = NULL;

static char * man_name = NULL;
static int man_name_size = 0;
static char filename[FILENAME_LENGTH] = {'s','a','v','e','0','0'};

static int filename_found = 0;
static int number_of_phenotypes_written = 0;

extern double avr_attractor;
extern double avr_diff;

extern setting_flags flags;


extern int nr_of_success;
extern int wanted_attractor;

extern int nr_of_changed;
extern int nr_of_unchanged;
extern int finished_step;
extern int max_step;
extern double mutated_bits_pr_individ;
extern int nr_of_elites;
extern double fitness_activation_ratio;
void write_deveopment_start();

void set_filename(char * new_name, int new_size){
/*    if (man_name != NULL){
        free(man_name);
    }*/
    filename_found = new_size > 0;
    man_name = new_name;
    man_name_size = new_size;
}

char * find_file_name(){
    if (man_name_size > 0){
        return man_name;
    }
    
    if (filename_found){
        return filename;
    }
    
    DIR *dir = NULL;
    struct dirent *ent;
    int current_biggest = 0;
    
    char * fn_ptr = NULL;
    char biggest[3] = {'0', '0', '\0'};
    
    if ((dir = opendir(FILEPATH) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            fn_ptr = ent->d_name;
            if (ent->d_namlen == FILENAME_LENGTH){
                
                char current_sequence[2] = {0};
                current_sequence[0] = fn_ptr[ent->d_namlen-2];
                current_sequence[1] = fn_ptr[ent->d_namlen-1];
                
                int size = (current_sequence[0] - 48) * 10 + (current_sequence[1] - 48) + 1;

                if (size > current_biggest){
                    current_biggest = size;
                    sprintf(biggest, "%0i", current_biggest);
                }
            }
        }
        if (biggest[1] != '\0'){
            filename[FILENAME_LENGTH-2] = biggest[0];
            filename[FILENAME_LENGTH-1] = biggest[1];
        }else{
            filename[FILENAME_LENGTH-2] = '0';
            filename[FILENAME_LENGTH-1] = biggest[0];
        }
        
        closedir(dir);
        return filename;
    }
    return NULL;
}

void open_file_with_flag(int tmp_file, char * flags){
    if (usage_file == NULL){
        if (tmp_file){
            char path_and_name[1024] = {0};
            strcpy(path_and_name, FILEPATH);
            strcat(path_and_name, TMP_FILENAME);
            usage_file = fopen(path_and_name, flags);
        } else if (man_name_size > 0){
            char path_and_name[1024] = {0};
            strcpy(path_and_name, FILEPATH);
            strcat(path_and_name, man_name);
            usage_file = fopen(path_and_name, flags);
            filename_found = usage_file != NULL;
        } else if (find_file_name() != NULL){
            char path_and_name[1024] = {0};
            strcpy(path_and_name, FILEPATH);
            strcat(path_and_name, filename);
            usage_file = fopen(path_and_name, flags);
            filename_found = usage_file != NULL;
        }
    }
}

void close_file(){
    fclose(usage_file);
    usage_file = NULL;
}

void clear_filename_buffer(){
    filename_found = 0;
}

void clear_tmp(){
    number_of_phenotypes_written = 0;
    open_file_with_flag(1, "w");
    write_deveopment_start();
    close_file();
}

fpos_t get_file_pos(){
    fpos_t pos;
    fgetpos(usage_file, &pos);
    return pos;
}

void open_save_file(int temp_file){
    open_file_with_flag(temp_file, "a");
}

void open_attractor_dev_file(){
    char path_and_name[1024] = {0};
    strcpy(path_and_name, FILEPATH);
    strcat(path_and_name, man_name);
    strcat(path_and_name, "_dev");
    usage_file = fopen(path_and_name, "a");
}
void close_attractor_dev_file(){
    close_file();
}

void close_save_file(int temp_file){
    close_file();
}

void open_read_file(int temp_file){
    open_file_with_flag(temp_file, "r");
}

void close_read_file(int temp_file){
    close_file();
}

void write_regulation_heading(int restriction){

    if (flags.use_zero_init){
        
    fprintf(usage_file, "\n********************************************\nWanted attractor: %i\nMutation rate %f\nFitness max for rules %f\nElites %i\nZero init\nRestriction %i\n********************************************\n", wanted_attractor, mutated_bits_pr_individ, fitness_activation_ratio, nr_of_elites, restriction);
    }else{
        fprintf(usage_file, "\n********************************************\nWanted attractor: %i\nMutation rate %f\nFitness max for rules %f\nElites %i\nRandom init\nRestriction %i\n********************************************\n", wanted_attractor, mutated_bits_pr_individ, fitness_activation_ratio, nr_of_elites, restriction);
    }
}

void write_param_results(int m_rate, int e_size, int f_index){
    fprintf(usage_file, "\nmutation rate %i\nelite size %i\nfitness_index %i\n", m_rate, e_size, f_index);
}

void write_success_rate(int number_total){
    fprintf(usage_file, "\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n    %i succesful of %i total\
            \nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n", nr_of_success, number_total);
}

void write_more_stats(){
    fprintf(usage_file, TAG_FLAGS_START);
    
    char * text = flags.use_exponential_adaptation ? "  Exponential_Adaptation\n" : "  Linear_Adaptation\n" ;
    fprintf(usage_file, "%s", text);

    text = flags.use_linear_fitness ? "  Linear_fitness\n" : "  Exponential_fitness" ;
    fprintf(usage_file, "%s", text);
    
    text = flags.use_mutation_rate_pr_chromosome ? "  Mutation_pr_chromosome\n" : "  Mutation_pr_individual\n" ;
    fprintf(usage_file, "%s", text);
    
    text = flags.use_zero_init ? "  Zero_init\n" : "  Random_init\n";
    fprintf(usage_file, "%s", text);
    
    text = flags.is_successfull ? "  Successful\n" : "  Unsuccessful";
    fprintf(usage_file, "%s", text);
    
    fprintf(usage_file, TAG_FLAGS_STOP);
}

void write_statistics(attractor * a){
    fprintf(usage_file, "\n\n");
    fprintf(usage_file, TAG_STATS_START);
    
    fprintf(usage_file, "    Wanted %04.4i\n",
            wanted_attractor);
    
    if (flags.is_successfull){
        fprintf(usage_file, "    Successful\n");
    }else{
        fprintf(usage_file, "    Unsuccessful\n");
    }
    
    fprintf(usage_file, "    Attractor: trans %05.5i, atr %05.5i, traj %05.5i\n",
            a->transient,
            a->attractor_length,
            a->trajectory);
    
    fprintf(usage_file, "    Found on step %04.4i of %04.4i\n",
            finished_step+1,
            max_step);
    
    fprintf(usage_file, "    Avrerage attractor %f, average diff %f\n",
            avr_attractor,
            avr_diff);
    
    fprintf(usage_file, "    Changed %05.5i, Unchanged %05.5i\n",
            nr_of_changed,
            nr_of_unchanged);
    
    fprintf(usage_file, TAG_STATS_STOP);
}

void write_growth(int * min, int * max, int length){
    fprintf(usage_file, TAG_GROWTH_STAR);
    for (int i = 0; i < length; i++){
        fprintf(usage_file, "[%x, %x]", min[i], max[i]);
    }
    fprintf(usage_file, "\n");
    fprintf(usage_file, TAG_GROWTH_STOP);
}

void write_attractor_development(weighted_element * we, int size){
    fprintf(usage_file, TAG_WE_START);
    for (int i  = 0; i < size; i++){
        //more compact writeout wich has [transient,attractorlength,trajectory]
        fprintf(usage_file, " [%05.5i,%05.5i,%05.5i]\n",
                we[i].atr.transient,
                we[i].atr.attractor_length,
                we[i].atr.trajectory);
        /*
        fprintf(usage_file, "    Step:%04.4i, trans %05.5i, atr %05.5i, traj %05.5i\n",
                i,
                we[i].atr.transient,
                we[i].atr.attractor_length,
                we[i].atr.trajectory);
         */
    }
    fprintf(usage_file, TAG_WE_STOP);
}

void write_genotype(list * gtype, int states){
    fprintf(usage_file, TAG_RULE_START);
    
    char genes[6] = {0};
    int bits_n = (int)ceil((double)states / 2.0);
    
    int mask = pow(2, bits_n) - 1;
    
    for (int i = 0; i < gtype->size; i++){

        for (int j = 0; j < 6; j++){
            genes[j] = (unsigned char) (gtype->array[i] >> (bits_n * j)) & mask;
        }

        fprintf(usage_file, "    R%02.2i:N%x,S%x,W%x,C%x,E%x,Nxt%x >%i<\n", i, genes[5], genes[4], genes[3], genes[2], genes[1], genes[0], gtype->used[i]);
    }
    fprintf(usage_file, TAG_RULE_STOP);
}

void write_phenotype(linked_buffer * ptype, int size, int width){
    fprintf(usage_file, "    %06.6lu|%05.5i|%04.4i|%04.4i|%x", ptype->step, ptype->key, size, width, ptype->buffer[0]);
    
    for(int i  = 1; i < size; i++ ){
        fprintf(usage_file, ",%x",ptype->buffer[i]);
    }
    fprintf(usage_file, "\n");
}

void write_deveopment_start(){
    fprintf(usage_file, TAG_DEV_START);
}




int set_read_poss_to_init(const fpos_t * current_pos){
    char buffer[32] = {0};
    int length = 32;
    fsetpos(usage_file, current_pos);
    while (fgets(buffer, length, usage_file) != NULL){
        if (strcmp(buffer, TAG_DEV_START) == 0){
            return 1;
        }
    }
    return 0;
}


static int offset = 11;
static int static_line_size = 23;
static int dynamic_line_size = GRID_SIZE*GRID_SIZE;

int get_buffers_from(long step){

#if TIMING == 1
    start_timeing();
#endif
    if (usage_file == NULL){
#if TIMING == 1
        save_timeing_to(IOR);
#endif
        return 0;
    }
    int line_lenght = static_line_size + dynamic_line_size + dynamic_line_size;
    long poss = offset + line_lenght * step;
    fseek(usage_file, poss, SEEK_SET);
    
    linked_buffer * l = get_fist_hd_buffer();
    
    int size = 0;
    int width = 0;
    unsigned int buffer = 0;
        
    int ret = fscanf(usage_file, "%lu|%hu|%d|%d|", &(l->step), &(l->key), &size, &width );
        
    if (ret <= 0){
#if TIMING == 1
        save_timeing_to(IOR);
#endif
        return 0;//number of buffers filled.
    }
    for (int i = 0; i < size; i++){
        if (!fscanf(usage_file, "%d", &buffer)){
            fscanf(usage_file, ",%d", &buffer);
        }
        l->buffer[i] = (char)buffer;
    }

#if TIMING == 1
    save_timeing_to(IOR);
#endif
    return 1;    
}


//return buffers filled
int fill_hd_buffer(const fpos_t * current_pos){

#if TIMING == 1
    start_timeing();
#endif

    
    if (usage_file == NULL){
#if TIMING == 1

        save_timeing_to(IOR);
#endif
        return 0;
    }
    
    fpos_t pos = 0;
    fgetpos(usage_file, &pos);
    
    if (pos == 0){
        set_read_poss_to_init(current_pos);
    }
    
    int buffer_number = get_number_of_hd_buffers();
    linked_buffer * l = get_fist_hd_buffer();
    
    int size = 0;
    int width = 0;
    unsigned int buffer = 0;

    for (int x = 0; x < buffer_number; x++){

        int ret = fscanf(usage_file, "%lu|%hu|%d|%d|", &(l->step), &(l->key), &size, &width );
        
        if (ret <= 0){
#if TIMING == 1
            save_timeing_to(IOR);
#endif
            return x;//number of buffers filled.
        }
        for (int i = 0; i < size; i++){
            if (!fscanf(usage_file, "%d", &buffer)){
                fscanf(usage_file, ",%d", &buffer);
            }
            l->buffer[i] = (char)buffer;
        }
        l = l->next;
    }

#if TIMING == 1
    save_timeing_to(IOR);
#endif
    return buffer_number;
}



