//
//  fitness_functions.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#ifndef fitness_function
#define fitness_function


#include <stdio.h>
#include <math.h>
#include "structs.h"
#include "stats.h"
#include "fitness_function.h"
#include "development.h"
#include "phenotype.h"
#include "visualization.h"

//Extern set

//Extern get
extern setting_flags flags;
extern double fitness_activation_ratio;
extern double genome_weight_ratio;

double get_rule_fitness(automata *  ca){

    double sum = 0;
    for (int i = 0; i < ca->gtype->size; i++){
        sum += ca->gtype->used[i];
    }
    return (sum / ca->gtype->size);
}

double get_attractor_fitness(attractor * a, int biggest_difference, int wanted_attractor, int relative_close){
    
    /* FITNESS 1/(e^(x/50))
     * gives a rise towards 1, the close one gets, the closer it rises.
     * The number 50 is to give a more relative aspect to "close"
     */
    
    if (a == NULL){
        return 0.0;
    }
    
    double difference = fabs( (wanted_attractor - a->attractor_length) );
    if (flags.use_linear_fitness){
        biggest_difference ++;
        return (double)FITNESS_LINEAR(difference, biggest_difference);
    }else{
        return (double)FITNESS(difference, relative_close);
    }

}

void develop_an_attractor(automata * ca, weighted_element * we){
    attractor * a = run_development(ca, MAX_DEVELOPMENT_STEPS, USE_TEMPFILE);
    //TODO find structure fitness at step N.
    if (a == NULL){
        we->atr.trajectory = MAX_DEVELOPMENT_STEPS;
        we->atr.attractor_length = MAX_DEVELOPMENT_STEPS;
        we->atr.transient = 0;
    }else{
        we->atr = *a;
    }
    free(a);
}

void calculate_attractor_and_rule_fitness(automata * ca, weighted_element * we, int biggest_difference, int wanted_atractor, int relative_close){
    we->attractor_fitness = get_attractor_fitness(&(we->atr), biggest_difference, wanted_atractor, relative_close);
    we->compound_fitness = get_rule_fitness(ca);
}



void calculate_compound_fitness(weighted_element * we, double attractor_divider, double rule_divider, double genome_weight){
    
    double attractor_weight =  attractor_divider==0 ? 0 : (we->attractor_fitness / attractor_divider) * (1 - fitness_activation_ratio - genome_weight_ratio);
    
    double rule_weight =  rule_divider==0 ? 0 : (we->compound_fitness / rule_divider) * fitness_activation_ratio;
    
    genome_weight *= genome_weight_ratio;
    
    we->compound_fitness = attractor_weight + rule_weight + genome_weight;
}

#endif
