//
//  File.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 4/25/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include "gplotilizer.h"
#include "old_data_handling.h"
#include "path_and_name.h"


#define AVERAGE_INDEX 2
#define RUNS 20

static int started = 0;
extern int wanted_attractor;
int maximum_evo_steps = -1;


void write_plot_settings(FILE * f, const char * data_path){
    fprintf(f, "set terminal png medium\n");
    fprintf(f, "set output \"graph.png\"\n");
    fprintf(f, "cd '%s'\n", data_path);
    fprintf(f, "set style data lines\n");
    fprintf(f, "set title 'Parameters for AL: %d Steps:%d'\n", wanted_attractor, maximum_evo_steps);
    fprintf(f, "set key bmargin center horizontal\n");
    fprintf(f, "set xrange [0:%d]\n", maximum_evo_steps);
    fprintf(f, "set yrange [0:%d]\n", (int)(wanted_attractor*1.1));
    fprintf(f, "set ylabel \"Attractor size\"\n");
    fprintf(f, "set xlabel \"Generation number\"\n");
    fprintf(f, "set grid ytics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    fprintf(f, "set grid xtics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
}

void write_first_plot_line(FILE * f, const char * fn, const char * title, int index){
    fprintf(f, "plot '%s' using 1:%d title '%s'", fn, index, title);    
}

void write_added_plot_line(FILE * f, const char * fn, const char * title, int index){
    fprintf(f, ", \\\n'%s' using 1:%d title '%s'", fn, index, title);
}
void write_fill_between(FILE * f, const char * fn, const char * title, int index_lowr, int index_higher){

    fprintf(f, "plot '%s' using 1:%d:%d title 'Genotype size' w filledcu lc rgb \"#00C200\" axis x1y2,'' using 1:%d lt -1 notitle axis x1y2, '' using 1:%d lt -1 notitle axis x1y2",fn, index_lowr, index_higher, index_lowr, index_higher);

    
}
/*plot '%s' u 1:%d:%d w filledcu,'' u 1:%d lt -1 notitle, '' u 1:%d lt -1 notitle*/


FILE * get_write_plot_file(const char * plot_name, const char * flag){
    char file_and_path_store[1024] = {0};
    path_and_name_plot_script(file_and_path_store, plot_name);
    return fopen(file_and_path_store, flag);
}

void gplotilize_file(const char * name, int counter){
    char file_and_path_read[1024] = {0};
    path_and_name_handle(file_and_path_read, name);

    char file_and_path_store[1024] = {0};
    path_and_name_plot_data(file_and_path_store, name); 
    
    int * we_archive [RUNS] = {0};
    int archive_length[RUNS] = {0};
    
    for (int i = 0; i < counter; i++){
        we_archive[i] = malloc(sizeof(int) * maximum_evo_steps);
    }
    
    int step = 0;
    int al = 0;
    int p = 0;
    
    FILE * read_file = fopen(file_and_path_read, "r");
    while( seek_for_tag(TAG_WE_START, read_file) ){
        step = 0;
        
        
        
        while ( (0 <  fscanf(read_file, " [%*d,%d,%*d]\n", &al ) ) ||
               (0 <  fscanf(read_file, " Step:%*d, trans %*d, atr %d, traj %*d\n", &al ))){
            we_archive[p][step++] = al;
            archive_length[p] ++;
        }
        p++;
    }
    fclose(read_file);
    

    int max_steps = maximum_evo_steps;
   /* for (int i = 0; i < RUNS; i++){
        max_steps = (int)MAX(max_steps, archive_length[i]);
    }
    */
    FILE * write_file = fopen(file_and_path_store, "w");
    fprintf(write_file, "#Step   Avrg");
    fprintf(write_file, "    %2.2d", 0);        
    for (int i = 1; i < counter; i++){
        fprintf(write_file, "      %2.2d", i);        
    }
    
    for (int i  = 0; i < max_steps; i++){
        double avrg = 0;
        int iter = 0;
        for (int j = 0; j < counter; j++){
            if (i < archive_length[j]){
                avrg += (double)we_archive[j][i];
                iter ++;
            }else{
                avrg += (double)we_archive[j][archive_length[j]-1];
                iter ++;
            }
        }
        avrg /= (double)iter;

        fprintf(write_file, "%4.4i    %4.4f", i, avrg);
        
        for (int j = 0; j < counter; j++){
            if (i < archive_length[j]){
                fprintf(write_file, "    %4.4i", we_archive[j][i]);
            }else{
                fprintf(write_file, "    %4.4i", we_archive[j][archive_length[j]-1]);
            }
        }
        
        fprintf(write_file, "\n");
    }
    
    fclose(write_file);
    
    for (int i  = 0; i < counter; i++){
        free(we_archive[i]);
    }
    
}

void get_attractor_and_steps(const char * name){
    char file_and_path_store[1024] = {0};
    char title[1024] = {0};
    substring_title(title, name);

    path_and_name_handle(file_and_path_store, title);

    FILE * f = fopen(file_and_path_store, "r");
    fscanf(f, " ");
    skip_lines(1, f);
    fscanf(f, " Wanted attractor: %d\n", &wanted_attractor);

    seek_for_tag(TAG_STATS_START, f);
    skip_lines(3, f);
    fscanf(f, " Found on step %*d of %d", &maximum_evo_steps);

    fclose(f);
    if (maximum_evo_steps == -1 || wanted_attractor == -1){
        printf("unable to read wanted attractor and max steps\nexiting...");
        exit(1);
    }
}

void start_plot_script(const char * plot_name, const char * name, int index){
    started = 1;
    char filename[1024] = {0};
    name_plot_data(filename, name);
    
    char title[1024] = {0};
    substring_title(title, name);

    FILE * write_file = get_write_plot_file(plot_name, "w");
    
    write_plot_settings(write_file, PATH_FOR_PLOT_DATA);
    write_first_plot_line(write_file, filename, title, index);
    
    fclose(write_file);    
}

void add_to_plot_script(const char * plot_name, const char * name, int index){
    char filename[1024] = {0};
    name_plot_data(filename, name);
    
    char title[1024] = {0};
    substring_title(title, name);
    
    FILE * append_file = get_write_plot_file(plot_name, "a");
    write_added_plot_line(append_file, filename, title, index);
    fclose(append_file);
}


void gplot_average_and_best(const char * plot_name, int best, const char * fn, ca_settings set, int counter){
    wanted_attractor = set.wanted_attractor;
    maximum_evo_steps = set.max_loops;
    
    gplotilize_file(fn, counter);
    if (!started){
        start_plot_script(plot_name, fn, AVERAGE_INDEX);
    }else {
        add_to_plot_script(plot_name, fn, AVERAGE_INDEX);
    }
    if (0 <= best){
        add_to_plot_script(plot_name, fn, best + 3);
    }
    
}



//tar utgangspunkt i at man har dataen
void gplot_rule_usage_development(const char * fn, attractor a, int nr){

    char plot_name[256] = {0};
    sprintf(plot_name, "%s_%d_devstep", base_name_of_file(fn), nr);
//    strcpy(plot_name, base_name_of_file(fn));
//    strcat(plot_name, "_devstep");
    
    char filename[256] = {0};
    name_usage_in_dev(filename, fn);
    
    FILE *f = get_write_plot_file(plot_name, "w");
    fprintf(f, "reset\n");
    fprintf(f, "set terminal png medium\n");
    fprintf(f, "set output \"%s.png\"\n", plot_name);
    fprintf(f, "cd '%s'\n", PATH_FOR_PRODUCTS);
    fprintf(f, "set style data lines\n");
    fprintf(f, "set title 'Rule usage on %s attractor %d trajectory %d'\n", fn, a.attractor_length, a.trajectory);
    fprintf(f, "set key bmargin center horizontal\n");
    fprintf(f, "set xrange [0:%d]\n", a.trajectory + a.attractor_length);
    fprintf(f, "set yrange [0:%d]\n", (int)(16));
    fprintf(f, "set ylabel \"Rules used\"\n");
    fprintf(f, "set xlabel \"Development step\"\n");
    fprintf(f, "set grid ytics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    fprintf(f, "set grid xtics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
        
    write_first_plot_line(f, filename, "Death",   2);
    write_added_plot_line(f, filename, "Life",    3);
    
    write_added_plot_line(f, filename, "Change",  4);
    write_added_plot_line(f, filename, "Nochange",5);
    
    write_added_plot_line(f, filename, "Mod",     6);
    write_added_plot_line(f, filename, "Total",   7);
    
    fclose(f);
    
    
    strcpy(plot_name, base_name_of_file(fn));
    strcat(plot_name, "_cumulative");
    
    name_cumul(filename, fn);
    
    f = get_write_plot_file(plot_name, "w");
    fprintf(f, "reset\n");    
    fprintf(f, "set terminal png medium\n");
    fprintf(f, "set output \"%s.png\"\n", plot_name);
    fprintf(f, "cd '%s'\n", PATH_FOR_PRODUCTS);
    fprintf(f, "set style data lines\n");
    fprintf(f, "set title 'Cumulative rule usage on %s attractor %d trajectory %d'\n", base_name_of_file(fn), a.attractor_length, a.trajectory);
    fprintf(f, "set key bmargin center horizontal\n");
    fprintf(f, "set xrange [0:%d]\n", a.trajectory);
    fprintf(f, "set yrange [0:%d]\n", (int)(900));
    fprintf(f, "set ylabel \"Rules used\"\n");
    fprintf(f, "set xlabel \"Development step\"\n");
    fprintf(f, "set grid ytics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    fprintf(f, "set grid xtics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    
    write_first_plot_line(f, filename, "Death",   2);
    write_added_plot_line(f, filename, "Life",    3);
    
    write_added_plot_line(f, filename, "Change",  4);
    write_added_plot_line(f, filename, "Nochange",5);
    
    write_added_plot_line(f, filename, "Mod",     6);
    write_added_plot_line(f, filename, "Total",   7);
    
    fclose(f);
    
}


//tar utgangspunkt i at man har dataen
void gplot_special_growth(const char * plot_name, const char * fn){
    
    //write settings
    FILE *f = get_write_plot_file(plot_name, "w");
//    write_plot_settings(f, PATH_FOR_PLOT_DATA);
    fprintf(f, "reset\n");
    fprintf(f, "set terminal png medium\n");
    fprintf(f, "set output \"%s.png\"\n", fn);
    fprintf(f, "cd '%s'\n", PATH_FOR_PLOT_DATA);
    fprintf(f, "set multiplot\n");
    
    fprintf(f, "set style data lines\n");
    fprintf(f, "set title 'Parameters for AL: %d Steps:%d'\n", wanted_attractor, maximum_evo_steps);
    fprintf(f, "set key bmargin center horizontal\n\n");
    fprintf(f, "#unset key\n");
    fprintf(f, "set xlabel \"Generation number\"\n");
    fprintf(f, "set ylabel \"Attractor size\"\n");
    fprintf(f, "set y2label \"Genotype size\"\n\n");

    fprintf(f, "set xrange [0:%d]\n", maximum_evo_steps);
    fprintf(f, "set yrange [0:%d]\n", (int)(wanted_attractor*1.1));
    fprintf(f, "set y2range [0:55]\n\n");
    
    fprintf(f, "set ytics nomirror tc rgb \"#0000FF\"\n");
    fprintf(f, "set y2tics nomirror tc rgb \"#00C200\"\n\n");
    
    fprintf(f, "set mxtics 2\n");
    fprintf(f, "set mytics 2\n\n");
    
    fprintf(f, "set grid xtics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    
    fprintf(f, "set style line 1 lt 1 lw 2 pt 3 linecolor rgb \"red\"\n");
    char data_file[256] = {0};
    name_plot_data(data_file, fn);
    
    //plot average
//    write_first_plot_line(f, data_file, fn, 2);
//    name_plot_data(data_file, fn);

    //some min max plotting
    name_growth(data_file, fn);
    write_fill_between(f, data_file, fn, 2, 3);

    name_plot_data(data_file, fn);
    fprintf(f, ",\\\n'%s' using 1:%d title '%s' w l lc rgb \"#0000FF\"\n", data_file, 2, fn);
    
    fprintf(f, "unset grid\n");
    fprintf(f, "set xlabel \" \"\n");
    fprintf(f, "set xtics format \" \"\n");
    fprintf(f, "set ylabel \" \"\n");
    fprintf(f, "set y2label \" \"\n");
    

    fprintf(f, "set grid ytics lt 0 lw 1 lc rgb \"#3D9EFF\"\n");
    fprintf(f, "plot 1/0\n");
    
    fprintf(f, "unset grid\n");
    fprintf(f, "set grid y2tics lt 0 lw 1 lc rgb \"#00FF00\"\n");
    fprintf(f, "plot 1/0\n");
    fprintf(f, "unset multiplot\n");
    
    
    fclose(f);
}

void individual_gplot_special_growth(const char * plot_name, const char * fn, const char * dat, int target){
    
    //write settings
    FILE *f = get_write_plot_file(plot_name, "w");
    //    write_plot_settings(f, PATH_FOR_PLOT_DATA);
    fprintf(f, "set terminal jpeg medium\n");
    fprintf(f, "set output \"%s.jpeg\"\n", fn);
    fprintf(f, "cd '%s'\n", PATH_FOR_PLOT_DATA);
    fprintf(f, "set style data lines\n");
    fprintf(f, "set title 'Parameters for AL: %d Steps:%d'\n", wanted_attractor, maximum_evo_steps);
    fprintf(f, "set key bmargin center horizontal\n");
    
//    fprintf(f, "set ytics 2 nomirror tc lt 1\n");
    fprintf(f, "set y2range [0:40]\n");
    fprintf(f, "set y2tics 2 nomirror tc lt 2\n");
    fprintf(f, "set y2label \"Genotype size\"\n");
    
    fprintf(f, "set xrange [0:%d]\n", maximum_evo_steps);
    fprintf(f, "set yrange [0:%d]\n", (int)(wanted_attractor*1.1));
    fprintf(f, "set style line 1 lt 1 lw 2 pt 3 linecolor rgb \"red\"\n");
    fprintf(f, "set ylabel \"Attractor size\"\n");
    fprintf(f, "set xlabel \"Generation number\"\n");
    fprintf(f, "set grid ytics lt 0 lw 1 lc rgb \"#55bbbb\"\n");
    fprintf(f, "set grid y2tics lt 0 lw 1 lc rgb \"#bb55bb\"\n");
    fprintf(f, "set grid xtics lt 0 lw 1 lc rgb \"#bbbbbb\"\n");
    char data_file[256] = {0};
    name_plot_data(data_file, fn);
    
    //plot average
    //    write_first_plot_line(f, data_file, fn, 2);
    //    name_plot_data(data_file, fn);
    
    //some min max plotting
    name_growth(data_file, fn);
    write_fill_between(f, data_file, fn, 2, 3);
    
    name_plot_data(data_file, dat);
    fprintf(f, ",\\\n'%s' using 1:%d title '%s' w l ls 1 axis x1y1", data_file, target, fn);
    fclose(f);
}


void gplot_process(){
    
    DIR *dir = NULL;
    struct dirent *ent;
    if ((dir = opendir(PATH_FOR_HANDLING) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            if (ent->d_namlen > 4){
                if (strcmp( ( ent->d_name + (ent->d_namlen - 3) ), "dev") == 0){
                    if (wanted_attractor == -1){
                        get_attractor_and_steps(ent->d_name);
                    }
                    
                    gplotilize_file(ent->d_name, RUNS);
                    if (!started){
                        start_plot_script(DEFAULT_PLOT_NAME, ent->d_name, AVERAGE_INDEX);
                    }else{
                        add_to_plot_script(DEFAULT_PLOT_NAME, ent->d_name, AVERAGE_INDEX);
                    }
                    
                }
            }
        }
    }
    closedir(dir);
}