//
//  mutation.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 4/29/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <math.h>
#include "mutation.h"
#include "stats.h"
#include "chance.h"
/**
 * MUTATION
 */

static double probability = 0.0;

//Extern get
extern setting_flags flags;
extern double mutated_bits_pr_individ;

#define ADAPTIVE_MUTATION 1


void calculate_probability(automata * ca){
    int bits_pr_individ = ca->gtype->size * (ca->nh_size + 1) * ceil((double)ca->val.states / 2.0);
    probability = (mutated_bits_pr_individ / bits_pr_individ);
}


//Adaptive makes the probability from 100% to 5% of the calculated probability.
//Reasently changed this from exponential to linear to try and make it work better on
//higher attractorlengths.
void calculate_adaptive_propability(automata * ca,  weighted_element * we){
    
    int bits_pr_individ = ca->gtype->size * (ca->nh_size + 1) * ceil((double)ca->val.states / 2.0);
    if (flags.use_mutation_rate_pr_chromosome){
        probability = mutated_bits_pr_individ / ( (ca->nh_size + 1) * ceil((double)ca->val.states / 2.0) );
    }else {
        probability = ((mutated_bits_pr_individ) / bits_pr_individ);
    }
    
    if (flags.use_exponential_adaptation)
        probability *=  pow(0.05, we->attractor_fitness);
    else    //try at doing it linear and not some division exponential shit
        probability *=  (1 - (0.95 * we->attractor_fitness) );
}



void mutate_new_population(int number_of_elites, population * pop){
    
    int bits_pr_state = (int) ceil( (double) pop->new_pop[0]->val.states / 2.0);
    int nh_size = pop->new_pop[0]->nh_size + 1;
    int number_of_used_bits = nh_size * bits_pr_state;
    
    uint32_t safety_filter [16] = {0};
    uint32_t mask_filter [16] = {0};
    
    for (int i = 0; i < nh_size; i++){
        uint32_t max_state = (uint32_t)pop->new_pop[0]->val.states;
        safety_filter[i] = max_state << (i * bits_pr_state);
        mask_filter[i] = ((uint32_t)pow(2, (bits_pr_state)) - 1) << (i * bits_pr_state);
    }
    
    for (int i  = number_of_elites; i < pop->size; i++){
#if ADAPTIVE_MUTATION == 1
        calculate_adaptive_propability(pop->new_pop[i], &pop->fitness[i]);
#else
        calculate_probability(pop->new_pop[i]);
#endif
        
        
        list * gtype = pop->new_pop[i]->gtype;
        
        for (int j = 0; j < pop->new_pop[i]->gtype->size; j++){
            uint32_t mutations = 0;                         //mutation filter with no mutations
            
            for (int m = 0; m < number_of_used_bits; m++){
                if (hit_on_probability(probability)){                     //check for mutation on single bit
                    mutations |= (1 << m);                  //add 1 to mutation filter if hit
                }
            }
            
            if (mutations == 0){                            //continue if no mutations
                continue;
            }
            
            //flip mutations
            uint32_t result = gtype->array[j];              //copy chromosom to a temp variable
            result = result ^ mutations;                    //mutate temp variable
            
            gtype->array[j] = 0;                            //set original chrom to 0 and build it from
                                                            //legal values form the mutated temp variable
            for (int m = 0; m < nh_size; m++){
                uint32_t mod_result = (result & mask_filter[m]) % safety_filter[m];//make each gene safe
                gtype->array[j] |= mod_result;                                     //add to original
            }
            
            if ((gtype->array[j] >> bits_pr_state) == 0){   // if all rules are 0, whole chromosone is 0
                gtype->array[j] = 0;
            }
        }
    }
}
