//
//  old_data_handling.c
//  Experiment
//
//  Created by Andreas on 5/1/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <math.h>
#include "old_data_handling.h"
#include "development.h"
#include "program_setup.h"
#include "genotype.h"
#include "cellular_automata.h"
#include "visualization.h"
#include "gplotilizer.h"
#include "path_and_name.h"
#include "program_setup.h"

static FILE * f = NULL;

static attractor local_attractor;

extern excess * storage;
extern int storage_size;
extern int maximum_evo_steps;
extern int wanted_attractor;

void open_read(const char * name){
    f = fopen(name, "r");
}

void close_handle(){
    fclose(f);
}

void open_save(const char * name){
    f = fopen(name, "w");
}

int seek_for_tag(const char * tag, FILE * file){
    char buffer[1024] = {0};
    do {
        if (fgets(buffer, 1024, file) == NULL){
            return 0;
        }
    }while(strcmp(tag, buffer));
    return 1;
}

void skip_lines(int nr_of_lines, FILE * file){
    char buffer[1024] = {0};
    for (int i = 0; i < nr_of_lines; i++){
        fgets(buffer, 1024, file);
    }
}

void save_growing_evolution(const char * fn, int *** arg , int * length, int counter){


    double ** avrg = malloc(sizeof(double*)*60000);
    for (int i = 0; i < 60000; i++){
        avrg[i] = malloc(sizeof(double) * 2);
    }
    int longest = 0;

    for (int i = 0; i < counter; i++){
        longest = (int)MAX(longest, length[i]);
    }
    
    for (int i = 0; i < counter; i++){
        for (int j = 0; j < longest; j++){
            if (j < length[i]){
                avrg[j][0] += arg[i][j][0];
                avrg[j][1] += arg[i][j][1];
            }else {
                avrg[j][0] += arg[i][length[i]-1][0];
                avrg[j][1] += arg[i][length[i]-1][1];
            }
        }
    }
    
    for (int i = 0; i < longest; i++){
        avrg[i][0] = avrg[i][0] / (double)counter;
        avrg[i][1] = avrg[i][1] / (double)counter;
    }
    
    char path_and_name [256] = {0};
    path_and_name_growth(path_and_name, fn);
    
    open_save(path_and_name);
    fprintf(f, "#Step    Min     Max\n");

    for (int i = 0; i < longest; i++){
        fprintf(f, "%d      %f      %f\n", i, avrg[i][0], avrg[i][1] );
    }

    
    
    close_handle();
    
    for (int i  = 0; i < 60000; i++){
        free(avrg[i]);
    }
    free(avrg);
}

void individual_save_growing_evolution(const char * fn, int *** arg , int * length, int counter){
    
    for (int j = 0; j < counter; j++){
    
        char name[256] = {0};
        char path_and_name [256] = {0};
        sprintf(name, "%s_%d",fn, j);

        path_and_name_growth(path_and_name, name);
    
        open_save(path_and_name);
        fprintf(f, "#Step    Min     Max\n");

        for (int i = 0; i < length[j]; i++){
            fprintf(f, "%d      %d      %d\n", i, arg[j][i][0], arg[j][i][1] );
        }
    
        close_handle();
    }
}


void growth_plots(const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
    
    int counter = 0;
    int length[20] = {0};
    
    int *** arg = malloc(sizeof(int**)*20);
    for (int i = 0; i < 20; i++){
        arg[i] = malloc(sizeof(int*)*60000);
        for (int j = 0; j < 60000; j++){
            arg[i][j] = malloc(sizeof(int)*2);
        }
    }
//    int arg[20][20000][2] = {0};
    
    while (seek_for_tag(TAG_STATS_START, f)) {     //found a run
        if (seek_for_tag(TAG_GROWTH_STAR, f)) {
            while (fscanf(f, " [%x, %x]", &arg[counter][length[counter]][0], &arg[counter][length[counter]][1]) == 2) {
                length[counter]++;
            }
            length[counter]--;
        }
        counter ++;
    }
    close_handle();
    
    save_growing_evolution(fn, arg, length, counter);
    
    
    for (int i = 0; i < 20; i++){
        for (int j = 0; j < 60000; j++){
            free(arg[i][j]);
        }
        free(arg[i]);
    }
    free(arg);
    
}

int individual_growth_plots(const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
    
    int counter = 0;
    int length[20] = {0};
    
    int *** arg = malloc(sizeof(int**)*20);
    for (int i = 0; i < 20; i++){
        arg[i] = malloc(sizeof(int*)*60000);
        for (int j = 0; j < 60000; j++){
            arg[i][j] = malloc(sizeof(int)*2);
        }
    }
    
    while (seek_for_tag(TAG_STATS_START, f)) {     //found a run
        if (seek_for_tag(TAG_GROWTH_STAR, f)) {
            while (fscanf(f, " [%x, %x]", &arg[counter][length[counter]][0], &arg[counter][length[counter]][1]) == 2) {
                length[counter]++;
            }
            length[counter]--;
        }
        counter ++;
    }
    close_handle();
    
    individual_save_growing_evolution(fn, arg, length, counter);
    
    
    for (int i = 0; i < 20; i++){
        for (int j = 0; j < 60000; j++){
            free(arg[i][j]);
        }
        free(arg[i]);
    }
    free(arg);
    return counter;
}






int get_number_of_indexes(const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    
    open_read(buffer);
    int counter = 0;
    while (seek_for_tag(TAG_STATS_START, f)) {counter ++;}
    close_handle();
    
    return counter;
}


void save_active(const char * fn,  int * active_array, int active_length, int* total_rulespace, double avrg_value[4]){
    char path_and_name [256] = {0};
    path_and_name_rule_usage(path_and_name, fn);
    open_save(path_and_name);
    fprintf(f, "#Wanted %d, type %s\n", local_attractor.attractor_length, base_name_of_file(fn));
    fprintf(f, "#Nr          Active       Total\n");

    for (int i = 0; i < active_length; i++){
        fprintf(f, "%03.3d          %03.3d          %03.3d\n", i, active_array[i], total_rulespace[i]);           

    }    
    fprintf(f, "Success on average: %.3f of %.3f\nTotal on average: %.3f of %.3f\n",
            avrg_value[SUCCESS_AVRG],
            avrg_value[SUCCESS_AVRG_FULL],
            avrg_value[TOTAL_AVRG],
            avrg_value[TOTAL_AVRG_FULL]);
    
    close_handle();
    
    
    path_and_name_histogram(path_and_name, fn);
    open_save(path_and_name);
    
    fprintf(f, "\
            #attractor  active  size\n\
            %d          %.3f      %.3f\n\
            %d          %.3f      %.3f\n",
            local_attractor.attractor_length,
            avrg_value[SUCCESS_AVRG],
            avrg_value[SUCCESS_AVRG_FULL] - avrg_value[SUCCESS_AVRG],
            local_attractor.attractor_length,
            avrg_value[TOTAL_AVRG],
            avrg_value[TOTAL_AVRG_FULL] - avrg_value[TOTAL_AVRG]);
    
    close_handle();
    
}


void active_rules_for_success(const char * filename){
    char buffer[256] = {0};
    path_and_name_handle(buffer, filename);
    open_read(buffer);
    
    int total_rulespace[128] = {0};
    int active_array[128] = {0};
    int active_length = 0;
    
    double avrg_values[4]= {0.0};
    int counter = 0;
    while (seek_for_tag(TAG_STATS_START, f)) {     //found a run
        skip_lines(1, f);        
        
        fpos_t pos_a = 0;
        fpos_t pos_b = 0;
        fscanf(f, " ");
        fgetpos(f, &pos_a);
        fscanf(f, "Successful");
        fgetpos(f, &pos_b);
        
        if (seek_for_tag(TAG_RULE_START, f)){  //point to start of rules
            int active = 0;
            int total = 0;
            int read = 0;
            while (0 < fscanf(f, " R%*d:N%*d,S%*d,W%*d,C%*d,E%*d,Nxt%*d >%d<", &read)){
                active += read;
                total ++;
            }

            avrg_values[TOTAL_AVRG] += active;
            avrg_values[TOTAL_AVRG_FULL] += total;
            counter ++;
            if (pos_a != pos_b){
                total_rulespace[active_length] = total;
                active_array[active_length++] = active;
                avrg_values[SUCCESS_AVRG] += active;
                avrg_values[SUCCESS_AVRG_FULL] += total;
            }
        }
    }

    close_handle();
    avrg_values[SUCCESS_AVRG_FULL] = avrg_values[SUCCESS_AVRG_FULL] / active_length;
    avrg_values[SUCCESS_AVRG] = avrg_values[SUCCESS_AVRG] / active_length;
    avrg_values[TOTAL_AVRG_FULL] = avrg_values[TOTAL_AVRG_FULL] / counter;
    avrg_values[TOTAL_AVRG] = avrg_values[TOTAL_AVRG] / counter;
    save_active(filename, active_array, active_length, total_rulespace, avrg_values);
    
}

void genotype_at_pos(int pos){
    
}

int find_closest_to_best_index(const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
    
    double best_rate = 5000.00;
    int best_index = -1;
    int index = 0;
    int wanted = 0;
    int attractor = 0;
    double average = 0;
    while (seek_for_tag(TAG_STATS_START, f)) {
        fscanf(f, "  Wanted %d\n", &wanted);
        skip_lines(1, f);
        fscanf(f, " Attractor: trans %*d, atr %d, traj %*d\n", &attractor);        
        skip_lines(1, f);
        fscanf(f, "  Avrerage attractor %*lf, average diff %lf\n", &average);        
        if ((double)(wanted - attractor) + average < best_rate){
            best_rate = (double)(wanted - attractor) + average ;
            best_index = index;
        }

        index ++;
    } 
    close_handle();
    return best_index;
}

int find_best_index(const char * fn, int * rate, int *success_array, int * success_size){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
        
    int best_rate = 9999999;
    int best_index = -1;
    int index = 0;
    while (seek_for_tag(TAG_STATS_START, f)) {
        skip_lines(1, f);
        char buffer[1024] = {0};
        fscanf(f, " ");
        fgets(buffer, 1024, f);                 //read line
        if (strcmp("Successful\n", buffer) == 0){
            fgets(buffer, 1024, f);                 //read line

            int succes_step = 0;
            fscanf(f, " Found on step %d of %*d", &succes_step);
            success_array[(*success_size)++] = index;
            if (succes_step < best_rate){
                best_rate = succes_step;
                best_index = index;
            }
        }    
        index ++;
    } 
    close_handle();
    
    if (*success_size > 0){
        printf("successfull elements :\n%d", success_array[0]);
        for (int i = 1; i < *success_size; i++){
            printf(", %d",success_array[i]);
        }
        int new_index = -1;
        printf("\nWhich do you want the devstep from?\n");
        scanf("%d", &new_index);
        if (new_index != -1){
            best_index = new_index;
        }
    }
    
    *rate = best_rate;
    return best_index;
}


void add_chromosome_to_genotype(chromosome c, list * gtype, int index){
    if (gtype->size <= index){
        grow_genotype(gtype);
    }
    //hardcoded nr of states pr gene
    int bits_pr_state = 2;
    uint32_t mask = pow(2, bits_pr_state) - 1;
    gtype->array[index] = 0;
    gtype->array[index] |= (c.next & mask);    
    gtype->array[index] |= (c.east & mask) << (1 * bits_pr_state);
    gtype->array[index] |= (c.center & mask) << (2 * bits_pr_state);
    gtype->array[index] |= (c.west & mask) << (3 * bits_pr_state);
    gtype->array[index] |= (c.south & mask) << (4 * bits_pr_state);
    gtype->array[index] |= (c.north & mask) << (5 * bits_pr_state);
}

list * genotype_at_index(int wanted_index, const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);

    int current_index = 0;
    while(seek_for_tag(TAG_RULE_START, f)){
        if (wanted_index == current_index) {
            break;
        }
        current_index ++;
    }
    list * l = init_empty_genotype(1);

    int index = 0;
    chromosome c;
    while ( 0 < fscanf(f, " R%*d:N%d,S%d,W%d,C%d,E%d,Nxt%d >%*d<", &c.north, &c.south, &c.west, &c.center, &c.east, &c.next)) {
        add_chromosome_to_genotype(c, l, index);
        index++;
    }
    
    close_handle();    
    return l;
}

attractor attractor_at_index(int wanted_index, const char * fn){
    char buffer[256] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
    
    int current_index = 0;
    while(seek_for_tag(TAG_STATS_START, f)){
        if (wanted_index == current_index) {
            break;
        }
        current_index ++;
    }
    skip_lines(2, f);
    
    attractor a; 
    fscanf(f, " Attractor: trans %d, atr %d, traj %d", &a.transient, &a.attractor_length, &a.trajectory);

    close_handle();
    
    return a;
}


ca_settings settings_from_file(const char * fn){
    char buffer[1024] = {0};
    path_and_name_handle(buffer, fn);
    open_read(buffer);
    
    ca_settings sett;
    sett.linear = 1;
    
    fscanf(f, " ********************************************\
           Wanted attractor: %d\
           Mutation rate %lf\
           Fitness max for rules %lf\
           Elites %d",
           &sett.wanted_attractor, 
           &sett.mutation, 
           &sett.fitnes_weight, 
           &sett.elites);
    fscanf(f, " ");

    fgets(buffer, 1024, f);
    sett.zero_init = strcmp(buffer, "Zero init\n") == 0 ? 1 : 0;
    
    fscanf(f, "Restriction %d", &sett.restriction);
    
    seek_for_tag(TAG_STATS_START, f);
    skip_lines(3, f);

    fscanf(f, " Found on step %*d of %d", &sett.max_loops);
    
    close_handle();
    
    return sett;
}

GENOMTYPE type_from_filename(const char * fn){
    if (fn[0] == 'g')
        return GROWING;
    if (fn[0] == 'r')
        return RESTRICTED;
    //default
    return FULL;
}

void save_data_gathering( const char * fn, attractor a, int best_index){
    char path_and_name[1024] = {0};
    path_and_name_usage_in_dev(path_and_name, fn);
    
    open_save(path_and_name);
    fprintf(f, "#Best Index:%d     Transient: %d  Attractor: %d  Trajectory: %d\n", best_index, a.transient, a.attractor_length, a.trajectory);
    fprintf(f, "#Step        Death        Life         Change       Nochange     Mod          Total\n");
    int limit = a.trajectory + a.attractor_length + 1 < storage_size ? a.trajectory + a.attractor_length + 1 : storage_size ;
    for (int i = 0; i < limit; i++){
    fprintf(f, "%04.4d         %03.3d          %03.3d          %03.3d          %03.3d          %03.3d          %03.3d\n",
            i, 
            storage[i].death,
            storage[i].life,
            storage[i].change,
            storage[i].unchange,
            storage[i].mods,
            storage[i].death + storage[i].life + storage[i].change + storage[i].mods );
    }
    close_handle();
    
    
    path_and_name_cumul(path_and_name, fn);
    open_save(path_and_name);
    fprintf(f, "#Best Index:%d     Transient: %d  Attractor: %d  Trajectory: %d\n", best_index, a.transient, a.attractor_length, a.trajectory);
    fprintf(f, "#Step        Death        Life         Change       Nochange     Mod          Total\n");
    int cumul[6] = {0};
    for (int i = 0; i < a.trajectory; i++){
        cumul[0] += storage[i].death;
        cumul[1] += storage[i].life;
        cumul[2] += storage[i].change;
        cumul[3] += storage[i].unchange;
        cumul[4] += storage[i].mods;
        cumul[5] += (storage[i].death + storage[i].life + storage[i].change + storage[i].unchange + storage[i].mods);
        fprintf(f, "%04.4d         %03.3d          %03.3d          %03.3d          %03.3d          %03.3d          %03.3d\n",
                i,
                cumul[0], 
                cumul[1],
                cumul[2],
                cumul[3],
                cumul[4],
                cumul[5]);
    }
    close_handle();
}

int develop_info_on_best_individual(const char * fn, ca_settings set, ca_values val){

    int best_time = 0;
    int buffer[256] = {0};
    int buffer_size = 0;
    int best_index = find_best_index(fn, &best_time, buffer, &buffer_size);
    if (best_index < 0){
        best_index = find_closest_to_best_index(fn);
    }
    local_attractor = attractor_at_index(best_index, fn);
    list * l = genotype_at_index(best_index, fn);
    set.type = type_from_filename(fn);
    
    automata * ca = init_automata(val);
    ca->gtype = l;
    
    develop_with_data_gathering(ca, set);

    free_automata(ca);
    
    save_data_gathering( fn, local_attractor, best_index);
    
    set_excessive(0);    
    
    return best_index;
}

void develop_info_on_successfull_individuals(const char * fn, ca_settings set, ca_values val){
    
    int best_time = 0;
    int buffer[256] = {0};
    int buffer_size = 0;
    char name[256] = {0};
    
    
    find_best_index(fn, &best_time, buffer, &buffer_size);
    double average_transient = 0.0;
    for (int i = 0; i < buffer_size; i++){
        
        local_attractor = attractor_at_index(buffer[i], fn);
        average_transient += local_attractor.transient;
        list * l = genotype_at_index(buffer[i], fn);
        set.type = type_from_filename(fn);
        
        automata * ca = init_automata(val);
        ca->gtype = l;
        
        develop_with_data_gathering(ca, set);
        
        free_automata(ca);
        
        sprintf(name, "%s_%d", fn, buffer[i]);
        save_data_gathering( name, local_attractor, buffer[i]);
        
        set_excessive(0);
        
        gplot_rule_usage_development(name, local_attractor, i);
    }
    
    FILE * fafa = fopen("/Users/andregis/transients.t", "a");
    
    fprintf(fafa, "%s:\n", fn);
    fprintf(fafa, "%.3f transient\n\n", average_transient/buffer_size);
    
    fclose(fafa);
}



int is_grower(const char * name){
    char prefix[256] = {0};
    strcpy(prefix, name);
    prefix[2] = 0;
    return !strcmp(prefix, "gr");
}

static int growth_counter = 0;

void handle_it(){
    
    DIR *dir = NULL;
    struct dirent *ent;
    if ((dir = opendir(PATH_FOR_HANDLING) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            
            if (strcmp( ( ent->d_name + (ent->d_namlen - 3) ), "dev") == 0){
                
                char name_without_dev[256] = {0};
                strcpy(name_without_dev, ent->d_name);
                name_without_dev[ent->d_namlen - 4] = 0;
                
                ca_settings set = settings_from_file(name_without_dev);
                //hardcoded values for now since they are not stored in runs.
                ca_values val = {.dim=2, .grid_size=16, .states=3};

                /**
                 * Find the active rules pr development step, also the amount of 
                 * rule types used at each step.
                 */
                int best_index = develop_info_on_best_individual(name_without_dev, set, val);
        
                
                /**
                 * Get number of active rules, and total rule pr success
                 * also get the average
                 */
                active_rules_for_success(name_without_dev);
                
                /**
                 * Creates plots with average and best individual
                 */
                best_index = -1;
                int number_of_indexex = get_number_of_indexes(name_without_dev);
                gplot_average_and_best("Avrg_and_best", best_index, ent->d_name, set, number_of_indexex);

                /**
                 * Create plots with the rule usage
                 */
                gplot_rule_usage_development(name_without_dev, local_attractor, 0);
                
                /**
                 * Creates the average growth in genomesize over generations
                 */
                if (is_grower(name_without_dev)){
                    char surfix[8] = {0};
                    char growth_name[256] = {0};
                    strcpy(growth_name, "Growth_with_genome_growth");
                    sprintf(surfix, "_%d", growth_counter++);
                    strcat(growth_name, surfix);
                    
                    growth_plots(name_without_dev);
                    gplot_special_growth(growth_name, name_without_dev);

                }
            }
        }
    }
}


void individual_plots(){
    
    DIR *dir = NULL;
    struct dirent *ent;
    if ((dir = opendir(PATH_FOR_HANDLING) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            
            if (strcmp( ( ent->d_name + (ent->d_namlen - 3) ), "dev") == 0){
                
                char name_without_dev[256] = {0};
                strcpy(name_without_dev, ent->d_name);
                name_without_dev[ent->d_namlen - 4] = 0;
                
                ca_settings set = settings_from_file(name_without_dev);
                wanted_attractor = set.wanted_attractor;
                maximum_evo_steps = set.max_loops;
                
                if (is_grower(name_without_dev)){
                    
                    gplot_average_and_best("Avrg_and_best", 0, ent->d_name, set, 20);
                    char growth_name[256] = {0};
                    growth_counter++;
                    int counter = individual_growth_plots(name_without_dev);

                    for (int x = 0; x < counter; x++){
                        char counter_name[256] = {0};
                        sprintf(growth_name, "%s_%d_%d","Growth_with_genome_growth",growth_counter, x);
                        sprintf(counter_name, "%s_%d", name_without_dev, x);
                        individual_gplot_special_growth(growth_name, counter_name, name_without_dev, x);

                    }
                }
            }
        }
    }
}




void rule_graphs_for_successful(){
    
    DIR *dir = NULL;
    struct dirent *ent;
    if ((dir = opendir(PATH_FOR_HANDLING) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            
            if (strcmp( ( ent->d_name + (ent->d_namlen - 3) ), "dev") == 0){
                
                char name_without_dev[256] = {0};
                strcpy(name_without_dev, ent->d_name);
                name_without_dev[ent->d_namlen - 4] = 0;
                
                ca_settings set = settings_from_file(name_without_dev);
                //hardcoded values for now since they are not stored in runs.
                ca_values val = {.dim=2, .grid_size=16, .states=3};
                
                develop_info_on_successfull_individuals(name_without_dev, set, val);

            }
        }
    }
}





void redevelop_in_handle(){
    DIR *dir = NULL;
    struct dirent *ent;
    if ((dir = opendir(PATH_FOR_HANDLING) ) != NULL ){
        while ((ent = readdir(dir)) != NULL){
            
            if (strcmp( ( ent->d_name + (ent->d_namlen - 3) ), "dev") == 0){
                
                char name_without_dev[256] = {0};
                strcpy(name_without_dev, ent->d_name);
                name_without_dev[ent->d_namlen - 4] = 0;
                
                redevelop_for_different_geometry(name_without_dev);
            }
        }
    }
}

