//
//  phenotype.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "phenotype.h"
#include "cellular_automata.h"
#include "file_helper.h"

static linked_buffer ** archive = NULL;


void free_ptype_mem_buffers(automata * ca){
    if (archive != NULL){
        for (int i = 0; i < ca->ptype->nr_buffs; i++){
            free(ca->ptype->buffers[i]->buffer);
            free(ca->ptype->buffers[i]);
        }
        free(ca->ptype->buffers);
        archive = NULL;
    }
}

ptype * init_empty_ptype(automata * ca){
    ptype * p = malloc(sizeof(ptype));

    p->buffer_size = ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size;
    p->buffer_width = ca->val.grid_size;

    return p;
}

void setup_ptype(ptype * p, int buffer_n){

    int buffer_size = p->buffer_size;
    
    if (archive == NULL){
        linked_buffer * prev_buff = NULL;
        archive = (linked_buffer**)malloc(sizeof(linked_buffer*) * buffer_n);
        
        for (int i = 0; i < buffer_n; i++){
            
            archive[i] = (linked_buffer*) malloc(sizeof(linked_buffer));
            archive[i]->buffer = (char*)calloc(buffer_size, sizeof(char));
            archive[i]->key = 0;
            if(prev_buff != NULL){
                prev_buff->next = archive[i];
            }
            prev_buff = archive[i];
        }
        if (prev_buff != NULL){
            prev_buff->next = archive[0];
        }
    }
    
    p->indic = 0;
    p->nr_buffs = buffer_n;
    
    p->buffers = archive;
    
    p->crnt = get_indicated_mem_buffer(p);
    p->nxt = p->crnt->next;
}

void save_number_mem_buffer_to_disk(automata * ca, int skip_amount, int save_amount, int temp_file){
    linked_buffer * l = get_indicated_mem_buffer(ca->ptype)->next;
    for (int i = 0; i < skip_amount; i++){
        l = l->next;
    }
    open_save_file(temp_file);
    for (int i = 0; i < save_amount; i++){
        write_phenotype(l, ca->ptype->buffer_size, ca->ptype->buffer_width);
        l = l->next;
    }
    close_save_file(temp_file);
}

void save_mem_buffer_to_disk(automata * ca, int temp_file){
    save_number_mem_buffer_to_disk(ca, 0, ca->ptype->nr_buffs, temp_file);
}

linked_buffer * get_indicated_mem_buffer(ptype * p){
    return p->buffers[p->indic];
}

void set_initial_condition_on_ptype(ptype * ptype){
    int height = (ptype->buffer_size / ptype->buffer_width);
    int zygote_poss = ((height / 2) * ptype->buffer_width) + (ptype->buffer_width / 2);
    for (int i = 0; i < ptype->buffer_size; i++){
        ptype->crnt->buffer[i] = (i == zygote_poss) ? (char)1 : (char)0;
    }
    
}

void shift_ptype_buffer(automata * ca){
    ca->ptype->crnt = ca->ptype->nxt;
    ca->ptype->nxt = ca->ptype->nxt->next;
    ca->ptype->indic++;
    ca->ptype->indic = ca->ptype->indic % ca->ptype->nr_buffs;
}


