//
//  program_setup.c
//  Experiment
//
//  Created by Andreas on 5/1/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <string.h>
#include "program_setup.h"
#include "experimet.h"
#include "development.h"
#include "phenotype.h"
#include "genotype.h"
#include "cellular_automata.h"
#include "attractor_checking.h"
#include "visualization.h"
#include "file_helper.h"
#include "mersenne_twister.h"
#include "old_data_handling.h"
#include "timeing.h"
#include "visualization.h"

extern setting_flags flags;
extern int rounds;

extern int nr_of_success;
extern int wanted_attractor;
extern int nr_of_changed;
extern int nr_of_unchanged;
extern double mutated_bits_pr_individ;
extern int nr_of_elites;
extern double fitness_activation_ratio;
extern int max_step;
extern double genome_weight_ratio;

extern int nr_of_states;
extern int size_of_grid;
extern int nr_of_dimensions;


void develop_with_data_gathering(automata * ca, ca_settings settings){
        
    
    //setting up the automata.
    int number_of_mem_buffs_pr_ca = MEM_BUFFERS;
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, number_of_mem_buffs_pr_ca);
    
    int phenotype_hd_buffer_size  = ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size;
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    
    set_excessive(1);
    
    attractor * a = run_development(ca, MAX_DEVELOPMENT_STEPS, 1);
   // visualize_development(16, 4);
    free(a);    
    free_ptype_mem_buffers(ca);    
    free_hd_buffers();
}

void set_default_settings(){
    flags.use_zero_init =                   1;
    flags.use_linear_fitness =              1;
    flags.use_exponential_adaptation =      0;
    flags.use_mutation_rate_pr_chromosome = 0;
    
    nr_of_elites =                      1;
    nr_of_success =                     0;

    mutated_bits_pr_individ =           2.6;
    fitness_activation_ratio =          0.0;
    genome_weight_ratio =               0.2;
    
    rounds =                            20;
    wanted_attractor =                  40;
    max_step =                          5000;
}

void save_header(int restriction){
    open_save_file(0);
    write_regulation_heading(restriction);
    close_save_file(0);
}

void save_successrate(){
    open_attractor_dev_file();
    write_success_rate(rounds);
    close_attractor_dev_file();
}

void test_longest_gen(){
    start_timeing_2();
    set_default_settings();

    wanted_attractor = 10000;
    mutated_bits_pr_individ = 1.8;
    
    //GROWING
    set_filename("long_long_time1000", 1);
    max_step = 1000;
    save_header(25);
        
    run_regulate_testing(1, GROWING, 25);
        
    save_successrate();
    nr_of_success = 0;
    printf("\n1000\n");
    print_current_spent();
    
    
    //Next
    
    start_timeing_2();
    set_filename("long_long_time2000", 1);
    max_step = 2000;
    save_header(25);
    
    run_regulate_testing(1, GROWING, 25);
    
    save_successrate();
    nr_of_success = 0;
    printf("\n2000\n");
    print_current_spent();
    
    
    //Next
    
    start_timeing_2();
    set_filename("long_long_time5000", 1);
    max_step = 5000;
    save_header(25);
    
    run_regulate_testing(1, GROWING, 25);
    
    save_successrate();
    nr_of_success = 0;
    
    printf("\n5000\n");
    print_current_spent();
    
    
    
    ///Next
    
    start_timeing_2();
    set_filename("long_long_time10000", 1);
    max_step = 10000;
    save_header(25);
    
    run_regulate_testing(1, GROWING, 25);
    
    save_successrate();
    nr_of_success = 0;
    printf("\n10000\n");
    print_current_spent();
}

void fine_tune_mutation_rate_expon(){
    start_timeing_2();
    set_default_settings();
    flags.use_exponential_adaptation = 1;
    int wa[2] = {40, 100};
    double mutation_rates[3] = {2.7, 2.8, 2.9};
    //GROWING

    char * grow_fn[6] = {"gr_th25_mr2.7_40","gr_th25_mr2.7_100","gr_th25_mr2.8_40","gr_th25_mr2.8_100", "gr_th25_mr2.9_40", "gr_th25_mr2.9_100"};
    int thresh[1] = {25};

    for (int i = 0; i < 6; i++){
            
        wanted_attractor = wa[i%2];
        mutated_bits_pr_individ = mutation_rates[i/2];
        set_filename(grow_fn[i], 1);
    
        save_header(thresh[0]);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, GROWING, thresh[0]);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
    
    //RESTRICTION
    char * rest_fn[6] = {"res_10_mr2.7_40","rest_20_mr2.7_100","rest_10_mr2.8_40","rest_20_mr2.8_100", "rest_10_mr2.9_40", "rest_20_mr2.9_100"};
    int restrictions[2] = {10, 20};
    
    for (int i = 0; i < 6; i++){

        wanted_attractor = wa[i%2];
        mutated_bits_pr_individ = mutation_rates[i/2];
        set_filename(rest_fn[i], 1);

        save_header(restrictions[i%2]);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, RESTRICTED, restrictions[i%2]);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
    

    /*
//FULL
    set_filename("full", 1);
    open_save_file(0);
    write_regulation_heading(0);
    close_save_file(0);
    
    init_genrand(GLOBAL_RANDOM_SEED);
    run_regulate_testing(rounds, FULL, 0);
    
    open_attractor_dev_file();
    write_success_rate(rounds);
    close_attractor_dev_file();
    
    nr_of_success = 0;

    print_current_spent();
    */
}


void create_comparison_lin(){
    start_timeing_2();
    set_default_settings();
    
    int wa[2] = {40, 100};
    double mutation_rates[3] = {1.6, 2.6, 3.6};
    //GROWING
    
    char * grow_fn[6] = {"gr_th25_mr1.6_40","gr_th25_mr1.6_100","gr_th25_mr2.6_40","gr_th25_mr2.6_100", "gr_th25_mr3.6_40", "gr_th25_mr3.6_100"};
    int thresh[1] = {25};
    
    for (int i = 0; i < 6; i++){
        
        wanted_attractor = wa[i%2];
        mutated_bits_pr_individ = mutation_rates[i/2];
        set_filename(grow_fn[i], 1);
        
        save_header(thresh[0]);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, GROWING, thresh[0]);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
    
    //RESTRICTION
    char * rest_fn[6] = {"res_10_mr1.6_40","rest_20_mr1.6_100","rest_10_mr2.6_40","rest_20_mr2.6_100", "rest_10_mr3.6_40", "rest_20_mr3.6_100"};
    int restrictions[2] = {10, 20};
    
    for (int i = 0; i < 6; i++){
        
        wanted_attractor = wa[i%2];
        mutated_bits_pr_individ = mutation_rates[i/2];
        set_filename(rest_fn[i], 1);
        
        save_header(restrictions[i%2]);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        run_regulate_testing(rounds, RESTRICTED, restrictions[i%2]);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
    
}

void check_fittness_weighting(int attractor){
    start_timeing_2();
    set_default_settings();
    wanted_attractor = attractor;
    max_step = 4500;
    flags.use_exponential_adaptation = 1;
    
    nr_of_elites =                      0;
    fitness_activation_ratio =          0.15;
    genome_weight_ratio =               0.0;

    double mut[3] = {1.6, 2.6, 3.6};

    char names[4][256] = {0};    
    char surfix[32] = {0};
    sprintf(surfix, "_al%d", attractor);

    //GROW
    char * grow_fn[3] = {"grow_adap_exp_mut1.6","grow_adap_exp_mut2.6","grow_adap_exp_mut3.6"};
    for (int i = 0; i < 3; i++){
        strcpy(names[i], grow_fn[i]);
        strcat(names[i], surfix);
    }
    

    int threshold = 25;
    for (int i = 1; i < 2; i++){
        mutated_bits_pr_individ = mut[i];
       // init_genrand(GLOBAL_RANDOM_SEED);
        
        set_filename(names[i], 1);
        save_header(threshold);

        run_regulate_testing(rounds, GROWING, threshold);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
    
    //RESTRICTION
    
    char * rest_fn[3] = {"rest_mut1.6","rest_mut2.6","rest_mut3.6"};
    for (int i = 0; i < 3; i++){
        strcpy(names[i], rest_fn[i]);
        strcat(names[i], surfix);
    }
    

    
    int restrictions = attractor < 100 ? 10 : 20;
    
    for (int i = 0; i < 3; i++){
        mutated_bits_pr_individ = mut[i];
       // init_genrand(GLOBAL_RANDOM_SEED);

        set_filename(names[i], 1);
        save_header(restrictions);        

        run_regulate_testing(rounds, RESTRICTED, restrictions);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
}


void long_run(int steps, double mutation, double activation){
    start_timeing_2();
    set_default_settings();
    rounds = 1;
    wanted_attractor = 5000;
    
    max_step = steps;
    mutated_bits_pr_individ = mutation;
    fitness_activation_ratio = activation;
    set_filename("LongLongTest", 1);
    
    save_header(25);
    
    init_genrand(GLOBAL_RANDOM_SEED);
    run_regulate_testing(rounds, GROWING, 25);
    
    save_successrate();
    nr_of_success = 0;
    
    print_current_spent();
}



void growing_4x4_3state_experiment_short(){
    start_timeing_2();
    set_default_settings();
/*  Need to find the generaion for 2560 and such...
    int attractor_lengths[5] = {10, 40, 160, 640, 2560};
    int generation_steps[5] = {5000, 5000, 5000, 6000, 10000};
 */
    int attractor_lengths[5] = {10, 40, 160};
    int generation_steps[5] = {5000, 5000, 5000};
    char * names[3] = {"grow_al10", "grow_al40", "grow_al160"};
    
    int threshold = 25;
    for (int i = 0; i < 3; i++){
        wanted_attractor = attractor_lengths[i];
        max_step = generation_steps[i];
        set_filename(names[i], 1);
        
        save_header(threshold);
        
        run_regulate_testing(rounds, GROWING, threshold);
        
        save_successrate();
        nr_of_success = 0;
        
        print_current_spent();
    }
}

void rest_4x4_3state_experiment_short(){
    start_timeing_2();
    set_default_settings();
    /*  Need to find the generaion for 2560 and such...
     int attractor_lengths[5] = {10, 40, 160, 640, 2560};
     int generation_steps[5] = {5000, 5000, 5000, 6000, 10000};
     */
    int attractor_lengths[5] = {10, 40, 160};
    int generation_steps[5] = {5000, 5000, 5000};
    
    char * names[3][3] = {
        {"rest_al10_rest8",     "rest_al10_rest9",      "rest_al10_rest10"}, 
        {"rest_al40_rest13",    "rest_al40_rest14",     "rest_al40_rest15"},
        {"rest_al160_rest19",   "rest_al160_rest20",    "rest_al160_rest21"}
    };
    
    int restrictions[3][3] = {
        { 8,  9, 10},
        {13, 14, 15},
        {19, 20, 21}
    };
    
    for (int i = 0; i < 3; i++){
        wanted_attractor = attractor_lengths[i];
        max_step = generation_steps[i];
        
        for (int j = 0; j < 3; j++){
            set_filename(names[i][j], 1);
            
            save_header(restrictions[i][j]);
            
            run_regulate_testing(rounds, GROWING, restrictions[i][j]);
            
            save_successrate();
            nr_of_success = 0;
            
            print_current_spent();
        }
    }
}

void growing_4x4_single_run_for_gen_length(int attractor){
    start_timeing_2();
    set_default_settings();
    rounds = 1;
    max_step = 20000;
    wanted_attractor = attractor;
    nr_of_elites = 1;
    flags.use_mutation_rate_pr_chromosome = 1;
    
    set_filename("gr_default_gen20000_el1_mutationprchromosome", 1);
    save_header(15);
    run_regulate_testing(rounds, GROWING, 15);
    save_successrate();
    nr_of_success = 0;
    print_current_spent();

}

void n_for_params(char * dest, int thres, GENOMTYPE type, int states, int grid){
    switch (type) {
        case RESTRICTED:
            sprintf(dest, "re_al%d_mu%2.2f_th%d_gs%d_st%d",  wanted_attractor, mutated_bits_pr_individ, thres, grid, states);
            break;
            
        case GROWING:
            sprintf(dest, "gr_al%d_mu%2.2f_th%d_gs%d_st%d",  wanted_attractor, mutated_bits_pr_individ, thres, grid, states);
            break;
            
        default:
            sprintf(dest, "fu_al%d_mu%2.2f_gs%d_st%d",  wanted_attractor, mutated_bits_pr_individ, grid, states);
            break;
    }
}

void name_for_params(char * dest, int thres, GENOMTYPE type){
    n_for_params(dest, thres, type, 0, 0);
}


void test_for_elitism(int pr_individ){
    start_timeing_2();
    set_default_settings();
    flags.use_mutation_rate_pr_chromosome = !pr_individ;
    max_step = 5000;
    fitness_activation_ratio = 0.0;
    rounds = 20;

    int threshhold = 10;
    char name[256] = {0};

    int attractors[3] = {40, 120, 1560};
//    int restrictions[2][2] = {{10,11},{18,19}};
    
    flags.use_mutation_rate_pr_chromosome = 1;
    rounds = 1;
   
    mutated_bits_pr_individ = 2.6;
    flags.use_mutation_rate_pr_chromosome = 0;
    
    for (int i  = 2; i < 3; i++){
        wanted_attractor = attractors[i];
        
        name_for_params(name, threshhold, FULL);
        strcat(name, "_full");
        set_filename(name, 1);
        
        save_header(threshhold);
        run_regulate_testing(rounds, FULL, threshhold );
        save_successrate();
        
        nr_of_success = 0;
        print_current_spent();
    }
}

void config_mutation_pr_chrom(){
    start_timeing_2();
    set_default_settings();
    rounds = 10;
    flags.use_mutation_rate_pr_chromosome = 1;
    
    char name[256] = {0};
    
    int elites[2] = {0, 1};
    int attractors[3] = {40, 200, 600};
    double mutations[4] = {0.5, 1.5, 2.5, 3.5};
    int thres[4] = {15, 25, 35, 45};    
    
    for (int i = 0; i < 4; i++){

        for (int x = 0; x < 3; x++) {
            for (int p = 0; p < 4; p++) {
                mutated_bits_pr_individ = mutations[i];
                nr_of_elites = elites[1];
                wanted_attractor = attractors[x];
                
                name_for_params(name, thres[p], GROWING);
                set_filename(name, 1);
                
                save_header(thres[p]);
                run_regulate_testing(rounds, GROWING, thres[p]);
                save_successrate();
                
                nr_of_success = 0;
                print_current_spent();
            }
        }
    }
}

void main_experiment_part1(){
    start_timeing_2();
    set_default_settings();
    
    char name[256] = {0};
    int threshold = 10;
    int attractors[8] = {400, 200, 160, 80, 40, 20, 10, 7};
    int diff_rounds[8] = {10000, 8000, 8000, 6000, 6000, 4000, 4000, 4000};
    /*
    for (int i = 2; i < 8; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        //Growing
        n_for_params(name, threshold, GROWING, nr_of_states, size_of_grid);
        set_filename(name, 1);
        
        save_header(threshold);
        run_regulate_testing(rounds, GROWING, threshold);
        save_successrate();
        
        nr_of_success = 0;
        print_current_spent();
        
        //Full
        n_for_params(name, threshold, FULL, nr_of_states, size_of_grid);
        set_filename(name, 1);
        
        save_header(threshold);
        run_regulate_testing(rounds, FULL, threshold);
        save_successrate();
        
        nr_of_success = 0;
        print_current_spent();
    }
    
    
    int states[3] = {5,6,7};
    
    for (int i = 4; i < 8; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        
        for (int x = 0; x < 3; x++){
            //Growing
            nr_of_states = states[x];
            n_for_params(name, threshold, GROWING, nr_of_states, size_of_grid);
            set_filename(name, 1);
            
            save_header(threshold);
            run_regulate_testing(rounds, GROWING, threshold);
            save_successrate();
            
            nr_of_success = 0;
            print_current_spent();
        }
    }
    
*/

    int sizes[3] = {6, 8, 16};
    
    for (int i = 4; i < 8; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        
        for (int x = 0; x < 3; x++){
            //Growing
            size_of_grid = sizes[x];
            n_for_params(name, threshold, GROWING, nr_of_states, size_of_grid);
            set_filename(name, 1);
            
            save_header(threshold);
            run_regulate_testing(rounds, GROWING, threshold);
            save_successrate();
            
            nr_of_success = 0;
            print_current_spent();
        }
    }
    
}

void main_experiment_part2(){
    start_timeing_2();
    set_default_settings();
    
    char name[256] = {0};
    int threshold[8][3] = {
        {30, 35, 40}, //400
        {21, 25, 29}, //200
        {20, 24, 28}, //160
        {12, 15, 18}, //80
        {10, 13, 16}, //40
        {8, 10, 12},  //20
        {11, 13, 15}, //10
        {12, 16, 20}};//7
    
    int attractors[8] = {400, 200, 160, 80, 40, 20, 10, 7};
    int diff_rounds[8] = {10000, 8000, 8000, 6000, 6000, 4000, 4000, 4000};
    
    for (int i = 0; i < 8; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        
        //Restricted
        for (int x = 0; x < 3; x++){
            n_for_params(name, threshold[i][x], RESTRICTED, nr_of_states, size_of_grid);
            set_filename(name, 1);
            
            save_header(threshold[i][x]);
            run_regulate_testing(rounds, RESTRICTED, threshold[i][x]);
            save_successrate();
            
            nr_of_success = 0;
            print_current_spent();
        }
    }
}

void main_experiment_part3(){
    //change the number of sites pr state.
    start_timeing_2();
    set_default_settings();
    char name[256] = {0};
    int threshold = 10;
    
    int attractors[4] = {40, 20, 10, 7};
    int diff_rounds[4] = {4000, 4000, 4000, 4000};
    int states[3] = {5,6,7};
    
    fitness_activation_ratio = 0.1;
    genome_weight_ratio = 0.1;
    
    for (int i = 0; i < 4; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        
        for (int x = 0; x < 3; x++){
            //Growing
            nr_of_states = states[x];
            n_for_params(name, threshold, GROWING, nr_of_states, size_of_grid);
            set_filename(name, 1);
            
            save_header(threshold);
            run_regulate_testing(rounds, GROWING, threshold);
            save_successrate();
            
            nr_of_success = 0;
            print_current_spent();
        }
    }
}

void main_experiment_part4(){
    //change the number of sites pr state.
    start_timeing_2();
    set_default_settings();
    
    char name[256] = {0};
    int threshold = 10;
    
    int attractors[4] = {40, 20, 10, 7};
    int diff_rounds[4] = {6000, 4000, 4000, 4000};
    int sizes[3] = {6, 8, 16};
    
    for (int i = 0; i < 4; i++){
        wanted_attractor = attractors[i];
        max_step = diff_rounds[i];
        
        for (int x = 0; x < 3; x++){
            //Growing
            size_of_grid = sizes[x];
            n_for_params(name, threshold, GROWING, nr_of_states, size_of_grid);
            set_filename(name, 1);
            
            save_header(threshold);
            run_regulate_testing(rounds, GROWING, threshold);
            save_successrate();
            
            nr_of_success = 0;
            print_current_spent();
        }
    }
}

void redevelop_for_different_geometry(const char *fn){

    ca_values vals = {.dim = 2, .grid_size = 4, .states = 3};
    int rate = 0;
    int buffer_size = 0;
    int buffer[256] = {0};
    int index = find_best_index(fn, &rate, buffer, &buffer_size);
    int phenotype_hd_buffer_size  = vals.dim == 1 ? vals.grid_size : vals.grid_size * vals.grid_size;
    
    for (int i  = 0; i < buffer_size; i++){
        index = buffer[i];

    // 4x4
    FILE * saver = fopen("/Users/andregis/redevelopemt.stats", "a");
    fprintf(saver, "%s --> %d\n", fn, index);
    fclose(saver);
    
    vals.grid_size = 4;
    phenotype_hd_buffer_size  = vals.dim == 1 ? vals.grid_size : vals.grid_size * vals.grid_size;
    automata * ca = init_automata(vals);

    ca->gtype = genotype_at_index(index, fn);;
    
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, MEM_BUFFERS);
    set_initial_condition_on_ptype(ca->ptype);
    
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);

    attractor * a = run_development(ca, 10000, USE_TEMPFILE);

    saver = fopen("/Users/andregis/redevelopemt.stats", "a");
    
    fprintf(saver, "gs 4, trans:%d, attractor:%d, traject:%d\n",a->transient, a->attractor_length, a->trajectory);
    fclose(saver);
    
    if (a != NULL){
        free(a);
    }
    
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free_hd_buffers();
    // 6x6
    vals.grid_size = 6;
    phenotype_hd_buffer_size  = vals.dim == 1 ? vals.grid_size : vals.grid_size * vals.grid_size;
    ca = init_automata(vals);
    
    ca->gtype = genotype_at_index(index, fn);;
    
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, MEM_BUFFERS);
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    set_initial_condition_on_ptype(ca->ptype);
    
    a = run_development(ca, 3000, USE_TEMPFILE);
    
    saver = fopen("/Users/andregis/redevelopemt.stats", "a");
    
    fprintf(saver, "gs 6, trans:%d, attractor:%d, traject:%d\n",a->transient, a->attractor_length, a->trajectory);
    fclose(saver);
    
    
    if (a != NULL){
        free(a);
    }
    
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free_hd_buffers();
    
    //8x8
    vals.grid_size = 8;
    phenotype_hd_buffer_size  = vals.dim == 1 ? vals.grid_size : vals.grid_size * vals.grid_size;
    ca = init_automata(vals);
    
    ca->gtype = genotype_at_index(index, fn);
    
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, MEM_BUFFERS);
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    set_initial_condition_on_ptype(ca->ptype);

    a = run_development(ca, 3000, USE_TEMPFILE);

    saver = fopen("/Users/andregis/redevelopemt.stats", "a");
    
    fprintf(saver, "gs 8, trans:%d, attractor:%d, traject:%d\n",a->transient, a->attractor_length, a->trajectory);
    fclose(saver);
    
    if (a != NULL){
        free(a);
    }
    
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free_hd_buffers();
    
    //16x16
    vals.grid_size = 16;
    phenotype_hd_buffer_size  = vals.dim == 1 ? vals.grid_size : vals.grid_size * vals.grid_size;
    ca = init_automata(vals);
    
    ca->gtype = genotype_at_index(index, fn);
    
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, MEM_BUFFERS);
    init_hd_buffers(HD_BUFFERS, phenotype_hd_buffer_size);
    set_initial_condition_on_ptype(ca->ptype);
    
    a = run_development(ca, 3000, USE_TEMPFILE);

    saver = fopen("/Users/andregis/redevelopemt.stats", "a");
    
    fprintf(saver, "gs 16, trans:%d, attractor:%d, traject:%d\n",a->transient, a->attractor_length, a->trajectory);
    fclose(saver);
    
    if (a != NULL){
        free(a);
    }
    
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free_hd_buffers();
    
    }
    
}



