//
//  structs.h
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/27/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#ifndef Experiment_structs_h
#define Experiment_structs_h
#include <stdlib.h>
#include <stdint.h>

#define USE_TEMPFILE 1

#define USE_ATTRACTOR_FORCE 1
#define TIMING 1
#define RANDOM -1
#define NEXT_VALUE 0

//Default settings
#define GRID_SIZE 4
#define STATES 3
#define DIMS 2
#define POPULATION_SIZE 100
#define ADD_LIMIT 243
#define GLOBAL_RANDOM_SEED 556677

//BUFFERSETTINGS
#define HD_BUFFERS 1500
#define MEM_BUFFERS 50

//Fitness
#define FITNESS_DEMAND 1.0
#define FITNESS_GRAIN 0.00001

#define FITNESS(DIF, REL){1 / pow(M_E, (DIF / REL))}
#define FITNESS_LINEAR(DIF, REL){(REL - DIF) / REL};

#define  RESTRICT_LOW 10
#define  RESTRICT_MED 30
#define  RESTRICT_HIG 60

#define RELATIVE_CLOSE_10 5
#define RELATIVE_CLOSE_40 10
#define RELATIVE_CLOSE_100 20
#define RELATIVE_CLOSE_200 30
#define RELATIVE_CLOSE_500 60
#define RELATIVE_CLOSE_1500 200

//Attractor
#define WANTED_ATTRACTOR_10 10
#define WANTED_ATTRACTOR_40 40
#define WANTED_ATTRACTOR_100 100
#define WANTED_ATTRACTOR_200 200
#define WANTED_ATTRACTOR_500 500
#define WANTED_ATTRACTOR_1500 1500

//Normal probability in percentage
#define RANDOM_GRANUALITY 100
#define CROSSOVER_RATE 0.02 //percent
#define ADD_CHROMOSOM_RATE 0.02 //percent

//Mutation
#define MUTATION_GRANUALITY 100000.0

//LENGTH
#define MAX_EVOLUTION_STEPS 1500
#define MAX_DEVELOPMENT_STEPS 1500

//IO
#define SURFIX_PLOT_DATA ".dat"
#define SURFIX_GROWTH ".growth"
#define SURFIX_RULE_USAGE ".txt"
#define SURFIX_HISTOGRAM ".hist"
#define SURFIX_PLOT_SCRIPTS ".plot"
#define SURFIX_RULE_DEV ".rule"
#define SURfIX_RULE_USAGE_CUMULATIVE ".cumul"

#define PATH_FOR_PLOT_SCRIPTS "/Users/andregis/"
#define PATH_FOR_PLOT_DATA PATH_FOR_PLOT_SCRIPTS "Documents/plots/"
#define PATH_FOR_HANDLING PATH_FOR_PLOT_SCRIPTS "Documents/handle/"
#define PATH_FOR_PRODUCTS PATH_FOR_PLOT_SCRIPTS "Documents/prod/"

#define FILENAME_LENGTH 6
#define FILEPATH PATH_FOR_PLOT_SCRIPTS"/projects/"
#define TMP_FILENAME "tmp.out"

#define DEFAULT_PLOT_NAME "gplotilized"


//Tags
#define TAG_GROWTH_STAR "//GROWTH_STR\n"
#define TAG_GROWTH_STOP "//GROWTH_END\n"

#define TAG_STATS_START "//STATSSTR\n"
#define TAG_STATS_STOP "//STATSEND\n"

#define TAG_FLAGS_START "//FLAGSSTR\n"
#define TAG_FLAGS_STOP "//FLAGSSTP\n"

#define TAG_RULE_START "//RULESTR\n"
#define TAG_RULE_STOP "//RULEEND\n"

#define TAG_WE_START "//ATR_DEV_STR\n"
#define TAG_WE_STOP "//ATR_DEV_END\n"

#define TAG_DEV_START "//DEVSTR\n"

//Utillity
#define MAX(A, B){ B < A ? A : B}
#define MIN(A, B){ A <= B ? A : B}


enum TIMETYPE{DEV = 0, ATR = 1, IOR = 2, IOW = 3, GA = 4};
typedef enum GENOMTYPE{GROWING, RESTRICTED, FULL}GENOMTYPE;
typedef enum RULETYPE{NOTUSED = 0, NOCHANGE, GROWTH, DEATH, CHANGE, MODULUS}RULETYPE;

typedef enum AVRG_SUM{
    SUCCESS_AVRG = 0,
    SUCCESS_AVRG_FULL = 1,
    TOTAL_AVRG = 2,
    TOTAL_AVRG_FULL = 3
}AVRG_SUM;

static uint16_t rol(uint16_t value, uint16_t bits){
    bits &= 15;
    return ((value << bits) | (value >> (16 - bits)));
}
static uint16_t ror(uint16_t value, uint16_t bits){
    bits &= 15;
    return ((value >> bits) | (value << (16 - bits)));
}

typedef union use_linear_fitness{
    uint32_t all_flags;
    struct{
        uint32_t use_linear_fitness                 : 1;
        uint32_t use_exponential_adaptation         : 1;
        uint32_t use_zero_init                      : 1;
        uint32_t is_successfull                     : 1;
        uint32_t use_mutation_rate_pr_chromosome    : 1;
    };
}setting_flags;

typedef union chrom_zip{
    uint32_t whole;
    struct {
        uint32_t flags  : 2;
        uint32_t north  : 5;
        uint32_t south  : 5;
        uint32_t west   : 5;
        uint32_t center : 5;
        uint32_t east   : 5;
        uint32_t next   : 5;
    };
}chrom_zip;



typedef struct chromosome{
    int north;
    int south;
    int west;
    int center;
    int east;
    int next;  
}chromosome;

typedef struct ca_settings{
    int wanted_attractor;
    int restriction;
    int zero_init;
    int linear;    
    int elites;
    int max_loops;
    double mutation;
    double fitnes_weight;
    GENOMTYPE type;
}ca_settings;


typedef struct ca_values{
    int grid_size;
    char dim;
    char states;
}ca_values;

typedef struct content{
    long step;
    struct content * nxt;
} content;

typedef struct bucket{
    content * contents;
}bucket;

typedef struct range{
    int to;
    int from;
    int length;
}range;

typedef struct excess {
    int death;
    int life;
    int change;
    int unchange;
    int mods;
}excess;

typedef struct checkable{
    double value;
    int index;
}checkable;

typedef struct list{
    int size;
    int mods;
    char * used;
    uint32_t* array;
}list;

typedef struct linked_buffer{
    char * buffer;
    uint16_t key;
    long step;
    struct linked_buffer * next;
}linked_buffer;

typedef struct ptype{
    int buffer_size;
    int buffer_width;
    int indic;
    int nr_buffs;
    linked_buffer ** buffers;
    linked_buffer * crnt;
    linked_buffer * nxt;
}ptype;

typedef struct attractor{
    int transient;
    int attractor_length;
    int trajectory;
}attractor;

typedef struct automata{
    ca_values val;
    int nh_size;
    list * gtype;
    ptype * ptype;

}automata;

typedef struct weighted_element{
    double compound_fitness;
    double attractor_fitness;
    double genome_fitness;
    range range;

    attractor atr;
}weighted_element;

typedef struct population{
    automata ** cur;
    automata ** new_pop;
    weighted_element * fitness;
    int size;
}population;

#endif
