//
//  testing.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/27/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include "testing.h"
#include "cellular_automata.h"
#include "genotype.h"
#include "phenotype.h"
#include "visualization.h"
#include "development.h"
#include "file_helper.h"
#include "attractor_checking.h"
#include "stats.h"
#include "mersenne_twister.h"

static automata * ca = NULL;

void test_cellular_automata_creation(ca_values values){
    ca = init_automata(values);
    
}

void test_genotype_creation(){
    
    automata * ca = init_automata( (ca_values){.grid_size=5, .dim=2, .states=3});
    list * genotype = init_full_genotype(ca, NEXT_VALUE);
    
    for (int i = 0; i < genotype->size; i++){
        visualize_gene(genotype->array[i], ca->val.states);
    }
    free_genotype(genotype);
    
    /*********************/
    
    genotype = init_restricted_genotype(10, ca->val.states, RANDOM);
    
    for (int i = 0; i < genotype->size; i++){
        visualize_gene(genotype->array[i], ca->val.states);
    }
    free_genotype(genotype);
}

void test_phenotype_creation(){
    automata * ca = init_automata((ca_values){.grid_size=5, .dim=2, .states=3});
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, 10);
    linked_buffer * ptype = get_indicated_mem_buffer(ca->ptype);


    visualize_phenotype_buffer(ptype, ca->ptype->buffer_size, ca->ptype->buffer_width);
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    
}

void test_wolfram_ca(int number){
    //simple testing;
    
    //Create CA
    automata * ca = init_automata((ca_values){.grid_size=7, .dim=1, .states=2});
    attractor * a = NULL;
    ca->ptype = init_empty_ptype(ca);
    //Create genotype and save it to file
    ca->gtype = wolfram_gtype(ca, number);
    save_genotype_to_disk(ca->gtype, ca->val.states);
    
    //Initiate phenotype buffer for mem and harddrive. not sure which one should be bigger...
    setup_ptype(ca->ptype, 20);
    init_hd_buffers(20, ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size);

    //Setup phenotype for CA
    set_initial_condition_on_ptype(ca->ptype);

    //init finished.

    //run development and return attractor.
    a = run_development(ca, 40, USE_TEMPFILE);
    
    //optional visualization, not working
    visualize_development(ca->ptype->buffer_size, ca->ptype->buffer_width, a->trajectory);

    //free all the allocated arrays

    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free(a);
    
    return;
    
    
    /*
    automata * ca = init_automata(5, 1, 2);
    attractor * a = NULL;
    ca->gtype = wolfram_gtype(ca, number);
    save_genotype_to_disk(ca->gtype, ca->states);
    //create phenotype
    init_ptype_mem_buffer(ca, 2);
    init_hd_buffers(2, ca->dimesions == 1 ? ca->size : ca->size * ca->size);
    
    ca->current_ptype = get_indicated_mem_buffer();
    ca->next_ptype = ca->current_ptype->next;
    
    set_initial_condition_on_ptype(ca->current_ptype);
   // save_number_mem_buffer_to_disk(1);
    
    visualize_phenotype(ca->current_ptype);
    while (1) {
        a = run_development(ca, 2);

        if (a != NULL){
            visualize_development();
            free(a);
            free_automata(ca);
            free_ptype_mem_buffers();
            return;
        }
    }
     */
}

void test_2d_borders(){
    automata * ca = init_automata((ca_values){.grid_size=7, .dim=1, .states=2});
    attractor * a = NULL;
    
    //GROWING NORTH
    //Init pheno buffers
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, 5);
    init_hd_buffers(5, ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size);
    
    //Setup initial condition and pointers
    set_initial_condition_on_ptype(ca->ptype);
    
    // init and save genotype
    ca->gtype = growing_north_2d(ca);
    save_genotype_to_disk(ca->gtype, ca->val.states);
    
    a = run_development(ca, 17, USE_TEMPFILE);
    
    save_statistics(a);
    visualize_development(ca->ptype->buffer_size, ca->ptype->buffer_width, a->trajectory);
    free_genotype(ca->gtype);
    free(a);
    //GROWING WESST
    // init and save genotype
    ca->gtype = growing_west_2d(ca);
    save_genotype_to_disk(ca->gtype, ca->val.states);
    
    a = run_development(ca, 17, USE_TEMPFILE);
    save_statistics(a);
    visualize_development(ca->ptype->buffer_size, ca->ptype->buffer_width, a->trajectory);

    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free(a);
}

void test_modulus_rule(){
    automata * ca = init_automata((ca_values){.grid_size=5, .dim=2, .states=3});
    attractor * a = NULL;
    
    //GROWING NORTH
    //Init pheno buffers
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, 5);
    init_hd_buffers(5, ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size);
    
    //Setup initial condition and pointers
    set_initial_condition_on_ptype(ca->ptype);
    
    // init and save genotype
    ca->gtype = init_empty_genotype(4);
    ca->gtype->array[0] = 0x401;   //1N 0S 0W 0C 0E 0N -> 01 00 00 00 00 01 ->  0x402
    ca->gtype->array[1] = 0x210;   //0N 2S 0W 1C 0E 0N -> 00 10 00 01 00 00 ->  0x210
    ca->gtype->array[2] = 0x401;   //1N 0S 0W 0C 0E 0N -> 01 00 00 00 00 01 ->  0x402
    ca->gtype->array[3] = 0x009;   //0N 0S 0W 0C 2E 1N -> 00 00 00 00 10 01 ->  0x009
    save_genotype_to_disk(ca->gtype, ca->val.states);
    
    a = run_development(ca, 17, USE_TEMPFILE);
    save_statistics(a);
    visualize_development(ca->ptype->buffer_size, ca->ptype->buffer_width, a->trajectory);
    
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free(a);
}

void test_development(){

    //create a ca
    automata * ca = init_automata((ca_values){.grid_size=5, .dim=2, .states=3});
    attractor * a = NULL;

    //create phenotype
    ca->ptype = init_empty_ptype(ca);
    setup_ptype(ca->ptype, 10);
    

    //create genotype
    ca->gtype = init_full_genotype(ca, RANDOM);
    save_genotype_to_disk(ca->gtype, ca->val.states);
    
    
    //setup inital conditions
    set_initial_condition_on_ptype(ca->ptype);
    

    //run development   
    //visualize_phenotype(ca->current_ptype);
    a = run_development(ca, 9, USE_TEMPFILE);
    write_statistics(a);
    
    visualize_development(ca->ptype->buffer_size, ca->ptype->buffer_width, a->trajectory);
    //free
    free_ptype_mem_buffers(ca);
    free_automata(ca);
    free(a);

    
}


void run_development_speed_test(){
    automata * ca = init_automata((ca_values){.grid_size=4, .dim=2, .states=3});
    attractor * a = NULL;
    ca->ptype = init_empty_ptype(ca);
    
    setup_ptype(ca->ptype, 50);
    init_hd_buffers(1000, ca->val.dim == 1 ? ca->val.grid_size : ca->val.grid_size * ca->val.grid_size);
    
    //Setup initial condition and pointers
    set_initial_condition_on_ptype(ca->ptype);
    
    // init and save genotype
    ca->gtype = init_full_genotype(ca, RANDOM);
    save_genotype_to_disk(ca->gtype, ca->val.states);
    

    struct timeval  tv1, tv2;
    gettimeofday(&tv1, NULL);
    
    a = run_development(ca, 10000, USE_TEMPFILE);

    gettimeofday(&tv2, NULL);
    
    printf ("Total time = %f seconds\n",
            (double) (tv2.tv_usec - tv1.tv_usec)/1000000 +
            (double) (tv2.tv_sec - tv1.tv_sec));
    
    if (a != NULL){
        free(a);
    }
    free_hd_buffers();
    free_ptype_mem_buffers(ca);
    free_automata(ca);
}

void test_mutation(){
    /*flip a bit anywhere in the gene*/
    list * gtype = init_restricted_genotype(100, 3, 0);
    for (int j  = 0; j < gtype->size; j++){
        int bits_pr_state = 2;
        int states = 3;
        int number_of_used_bits = 12;
        int shift = genrand_int32() % number_of_used_bits;
        int mask = 3;
        uint32_t result = gtype->array[j] ^ (1 << shift); //flip sigle bit
        char x = mask & (result >> 6);
        if (x == 3){
            printf("\n");
        }
        int shift_value = shift / bits_pr_state;
        shift_value = shift_value*bits_pr_state;
        int potential_overflow = ((result >> shift_value) & mask);
        if ((states - 1) < potential_overflow ){
            potential_overflow = fmod(potential_overflow, (states - 1));

            result = ( result & ~(mask << shift_value) ) | ( potential_overflow << shift_value );
        }
        x = mask & (result >> 6);
        if (x == 3){
            printf("\n");
        }
        
        gtype->array[j] = result;
    }
    save_genotype_to_disk(gtype, 3);
    
    free_genotype(gtype);

}

void test_visualization();


