//
//  tweaking.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 4/29/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <string.h>
#include "tweaking.h"

#include "file_helper.h"
#include "experimet.h"
#include "mersenne_twister.h"


extern setting_flags flags;
extern int rounds;


extern int nr_of_success;
extern int wanted_attractor;
extern int nr_of_changed;
extern int nr_of_unchanged;
extern double mutated_bits_pr_individ;
extern int nr_of_elites;
extern double fitness_activation_ratio;

void zero_init_restricted_linear(){
    rounds = 20;                        //Number of rounds pr setting
    flags.use_zero_init = 1;                  //Genotype initiation type
    flags.use_linear_fitness = 1;
    nr_of_elites = 0;                   //Number of elites
    mutated_bits_pr_individ = 0.6;
    fitness_activation_ratio = 0.2;       //Starting with 0 weighting on active rules
    wanted_attractor = WANTED_ATTRACTOR_40;
    
    
    set_filename("restricted_linear_zero_init", 1);
    
    open_save_file(0);
    write_regulation_heading(RESTRICT_LOW);
    close_save_file(0);
    
    init_genrand(GLOBAL_RANDOM_SEED);
    
    run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW);
    
    
    open_attractor_dev_file();
    write_success_rate(rounds);
    close_attractor_dev_file();
}


void zero_init_restricted_exponential(){
    rounds = 20;                        //Number of rounds pr setting
    flags.use_zero_init = 1;                  //Genotype initiation type
    flags.use_linear_fitness = 0;
    nr_of_elites = 0;                   //Number of elites
    mutated_bits_pr_individ = 0.6;
    fitness_activation_ratio = 0.05;       //Starting with 0 weighting on active rules
    wanted_attractor = WANTED_ATTRACTOR_40;
    
    
    set_filename("restricted_exponential_zero_init", 1);
    
    open_save_file(0);
    write_regulation_heading(RESTRICT_LOW);
    close_save_file(0);
    
    init_genrand(GLOBAL_RANDOM_SEED);
    
    run_regulate_testing(rounds, RESTRICTED, RESTRICT_LOW);
    
    
    open_attractor_dev_file();
    write_success_rate(rounds);
    close_attractor_dev_file();
}


void growth_treshhold_configure(const char * surfix ){
    
    int treshhold[1] = {10};
    
    char * filenames[1];
    char fn1[256] = "gr_tr10";
    strcat(fn1, surfix);
    filenames[0] = fn1;
    
    for (int i = 0; i < 1; i++){
        set_filename(filenames[i], 1);
        nr_of_success = 0;
        
        open_save_file(0);
        write_regulation_heading(RESTRICT_LOW);
        close_save_file(0);
        
        init_genrand(GLOBAL_RANDOM_SEED);
        
        run_regulate_testing(rounds, GROWING, treshhold[i]);
        
        open_attractor_dev_file();
        write_success_rate(rounds);
        close_attractor_dev_file();
    }
}

void growth_mutation_configure(const char * surfix){
    double mutation[1] = {0.8};
    char * fix[1];
    char fn1[256] = "_mut0.8";
    strcat(fn1, surfix);
    fix[0] = fn1;
    

    
    for (int i = 0; i < 1; i++){
        mutated_bits_pr_individ = mutation[i];
        growth_treshhold_configure(fix[i]);
    }
}


void growth_weight_configure(const char * surfix){
    double weight[1] = {0.1}; //linear setting
    char * fix[1];
    char fn1[256] = "_we0.1";
    strcat(fn1, surfix);
    fix[0] = fn1;

    
    for (int i = 0; i < 1; i++){
        fitness_activation_ratio = weight[i];
        growth_mutation_configure(fix[i]);
    }
}

void growth_configure(){
    
    wanted_attractor = WANTED_ATTRACTOR_40;
    rounds = 20;                        //Number of rounds pr setting
    flags.use_zero_init = 1;                  //Genotype initiation type
    nr_of_elites = 0;                   //Number of elites
    
//    flags.use_linear_fitness = 1;
//    growth_weight_configure("_lin");
    
    flags.use_linear_fitness = 0;
    growth_weight_configure("_exp");
}