//
//  visualization.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/27/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "attractor_checking.h"
#include "visualization.h"
#include "file_helper.h"

extern int nr_of_states;

void print_seperator(){
    //printf("\033[2J");
    printf("\n");
}

void visualize_phenotype(linked_buffer * ptype, int size, int width){
    usleep(100000);
    if (size > width){
        print_seperator();        
    }
    for (int i = 0; i < size; i++){
        printf("%x", ptype->buffer[i]);
        if (fmod(i + 1, width) == 0){
            printf("\n");
        }
    }
    
}

void visualize_phenotype_buffer(linked_buffer *ptype, int size, int width){
    linked_buffer * visualize = ptype->next;

    do{
        visualize_phenotype(visualize, size, width);
        visualize = visualize->next;
        if (ptype == visualize)
            break;
    }while(1);
}



void visualize_hd_buffers(int number_to_vizualise, int size, int width){
    linked_buffer * l = get_fist_hd_buffer();
    int amount = number_to_vizualise < get_number_of_hd_buffers() ? number_to_vizualise : get_number_of_hd_buffers() ;
    
    for (int i = 0; i < amount; i++){
        visualize_phenotype(l, size, width);
        l = l->next;
    }
}

void visualize_development(int size, int width, int steps){
    open_read_file(USE_TEMPFILE);
    int buffers_filled = 0;
    fpos_t pos = 0;
    while (((buffers_filled = fill_hd_buffer(&pos) ) > 0 ) &&  0 < steps ){
        visualize_hd_buffers(buffers_filled, size, width);
        steps--;
    }
    close_read_file(USE_TEMPFILE);
}

void visualize_gene(uint32_t gene, int state_n){
    
    unsigned char genes[6] = {0};
    int bits_n = (int)ceil((double)state_n / 2.0);
    
    int mask = pow(2, bits_n) - 1;
    
    for (int i = 0; i < 6; i++){
        genes[i] = (unsigned char) (gene >> (bits_n * i)) & mask;
    }
    
    printf("Gene: North %x,  South %x,  West %x,  Center %x,  East %x,  Next %x  \n", genes[5], genes[4], genes[3], genes[2], genes[1], genes[0]);
 
}

void visualize_genotype(list * genotype){
    for (int i = 0; i < genotype->size; i++){
        visualize_gene(genotype->array[i], nr_of_states);
    }
}